# Ingestion related functions which may be overriden by an implementing app,
# e.g. Foursight or CGAP portal, using the dcicutils project_utils mechanism.

class SnovaultProjectIngestion:

    def note_ingestion_enqueue_uuids_for_request(self, ingestion_type, request, uuids):
        pass

    def note_submit_for_ingestion(self, submission_uuid, context):
        pass

    def note_post_ingestion(self, message, context):
        pass



# Loadxl related functions which may be overriden by an implementing app,
# e.g. Foursight or CGAP portal, using the dcicutils project_utils mechanism.

class SnovaultProjectLoadxl:
    def loadxl_order(self):
        return []



from collections import OrderedDict
from itertools import chain
from typing import Any, Dict, List
from urllib.parse import urlparse

from pyramid.httpexceptions import HTTPNotFound
from pyramid.view import view_config

from .etag import etag_app_version_effective_principals
from .interfaces import (
    COLLECTIONS,
    TYPES,
)
from .util import debug_log
from dcicutils.schema_utils import (
    SchemaConstants as sc,
    get_dependent_required,
    get_properties,
    get_required,
    get_items,
    is_array_schema,
    is_object_schema,
    is_submitter_required,
)
from .project_app import app_project


class SubmissionSchemaConstants:

    ENDPOINT = "/submission-schemas/"

    ALSO_REQUIRES = "also_requires"
    IS_REQUIRED = "is_required"
    PROHIBITED_IF_ONE_OF = "prohibited_if_one_of"
    REQUIRED_IF_NOT_ONE_OF = "required_if_not_one_of"


def includeme(config):
    config.add_route('schemas', '/profiles/')
    config.add_route('schema', '/profiles/{type_name}.json')
    config.add_route('submittables', SubmissionSchemaConstants.ENDPOINT)
    config.add_route(
        'submittable', SubmissionSchemaConstants.ENDPOINT + '{type_name}.json'
    )
    config.scan(__name__)


def _annotated_schema(type_info, request):
    """
    Add some extra annotiation to a schema obtained through given TypeInfo.
    Specifically, adds links to the /terms/ page and inheritance information
    through the `children` and `rdfs:subClassOf` properties. Also flags fields
    by write permission with the `readonly` field, if applicable.
    TODO: add flagging for user restricted fields once role-based field viewing
    is implemented.

    Args:
        type_info (TypeInfo): for an item type. See snovault.type_info.py
        request (Request): the current Request

    Returns:
        dict: the annotated schema
    """
    schema = type_info.schema.copy()
    schema['@type'] = ['JSONSchema']
    jsonld_base = request.registry.settings['snovault.jsonld.terms_namespace']
    schema['rdfs:seeAlso'] = urlparse(jsonld_base).path + type_info.name
    # add links to profiles of children schemas
    schema['children'] = [
        '/profiles/' + t_name + '.json' for t_name in type_info.child_types
    ]

    if type_info.factory is None:
        return schema

    # use first base_type that is not this type itself to handle abstract
    found_subtype = None
    for subtype in type_info.base_types:
        if subtype != type_info.name:
            found_subtype = subtype
            break
    if found_subtype:
        schema['rdfs:subClassOf'] = '/profiles/' + found_subtype + '.json'
    # add abstract flag to know if the profile represents abstract item
    schema['isAbstract'] = type_info.is_abstract

    collection = request.registry[COLLECTIONS][type_info.name]
    properties = OrderedDict()
    # add a 'readonly' flag to fields that the current user cannot write
    for k, v in schema['properties'].items():
        if 'permission' in v:
            if not request.has_permission(v['permission'], collection):
                v = v.copy()
                v['readonly'] = True
        properties[k] = v
    schema['properties'] = properties
    return schema


@view_config(route_name='schema', request_method='GET',
             decorator=etag_app_version_effective_principals)
@debug_log
def schema(context, request):
    """
    /profiles/{type_name}.json -- view for the profile of a specific item type
    A bit inefficient, but need to use the TypeInfo (not AbstractTypeInfo)
    to get the correct schema. To do this, iterate through all registered
    types until we find the one with matching item_type (given by type_name).
    This allows this endpoint to work with item name (e.g. MyItem) or item_type
    (e.g. my_item)
    """
    type_name = request.matchdict['type_name']
    types = request.registry[TYPES]
    found_type_info = None
    all_item_types = chain(types.by_item_type.values(),
                           types.by_abstract_type.values())
    for type_info in all_item_types:
        # handle both item name and item type inputs to the route (both valid)
        if type_info.name == type_name or type_info.item_type == type_name:
            found_type_info = type_info
            break
    if found_type_info is None:
        raise HTTPNotFound(type_name)
    return _annotated_schema(type_info, request)


@view_config(route_name='schemas', request_method='GET',
             decorator=etag_app_version_effective_principals)
@debug_log
def schemas(context, request):
    """
    /profiles/ view for viewing all schemas. Leverages the TypeInfo objects
    for regular classes using registry[TYPES].by_item_type and for abstract
    classes by using registry[TYPES].by_abstract_type
    """
    types = request.registry[TYPES]
    schemas = {}
    all_item_types = chain(types.by_item_type.values(),
                           types.by_abstract_type.values())
    for type_info in all_item_types:
        name = type_info.name
        schemas[name] = _annotated_schema(type_info, request)
    return schemas


def _get_conditionally_required_propnames(schema, keyword):
    """
    helper to look for the oneOf/anyOf declarations in the schema -
    for required properties and return a list of those conditionally
    required properties
    """
    propnames = []
    keyword_info = schema.get(keyword, [])
    for info in keyword_info:
        # is a list but examples so far have only one member?
        propnames.extend(info.get(sc.REQUIRED, []))
    return propnames


def _has_property_attr_with_val(propinfo, attrs_to_chk):
    """
    given a property with it's attributes will check against
    a dictionary of attribute names and values and returns true
    if the property has any of the attributes name/values in the dict
    Note: also works if the attr value is a list
    """
    for aname, avalues in attrs_to_chk.items():
        if aname in propinfo:
            propval = propinfo.get(aname)
            if isinstance(propval, list):
                if [pval for pval in propval if pval in avalues]:
                    return True
            elif propval in avalues:
                return True
    return False


def _get_item_name_from_schema_id(schema_id):
    """ this assumes a validly formatted $id value from a schema"""
    return schema_id.replace('/profiles/', '').replace('.json', '')


def _is_submittable_schema(schema_id, schema):
    """
    helper to determine if the schema has potentially submittable fields
    exclude abstract schemas by default
    """
    if schema.get('isAbstract'):
        return False
    # an explicit list of submittable items may be provided
    item_list = app_project().get_submittable_item_names()
    # a property in a schema that indicates this schema is submittable
    # eg. submmitter_id
    key_prop = app_project().get_prop_for_submittable_items()

    if item_list:
        if schema_id:
            item_name = _get_item_name_from_schema_id(schema_id)
            if item_name in item_list:
                return True  # explicitly named item found
    schema_props = get_properties(schema)
    if key_prop:
        prop_names = schema_props.keys()
        if key_prop in prop_names:
            return True  # the property that designates a schema as submittable was found

    return False


def _annotate_submittable_props(schema, props):
    """
    add annotations for requirements and dependencies to
    submittable props based on the schema info
    """
    required_props = get_required(schema)
    oneof_props = _get_conditionally_required_propnames(schema, sc.ONE_OF)
    anyof_props = _get_conditionally_required_propnames(schema, sc.ANY_OF)
    req_deps = get_dependent_required(schema)

    for propname, propinfo in props.items():
        _update_required_annotation(propname, propinfo, required_props)
        if propname in oneof_props:
            lprops = [p for p in oneof_props if p != propname]
            propinfo.setdefault(
                SubmissionSchemaConstants.REQUIRED_IF_NOT_ONE_OF, []
            ).extend(lprops)
            propinfo[SubmissionSchemaConstants.PROHIBITED_IF_ONE_OF] = lprops
        if propname in anyof_props:
            lprops = [p for p in anyof_props if p != propname]
            propinfo.setdefault(
                SubmissionSchemaConstants.REQUIRED_IF_NOT_ONE_OF, []
            ).extend(lprops)
        if propname in req_deps:
            propinfo[SubmissionSchemaConstants.ALSO_REQUIRES] = req_deps[propname]
    return props


def _update_required_annotation(
    property_: str,
    property_schema: Dict[str, Any],
    required_properties: List[str],
) -> None:
    """Add required annotation to property schema if appropriate."""
    if property_ in required_properties or is_submitter_required(property_schema):
        property_schema[SubmissionSchemaConstants.IS_REQUIRED] = True


def _build_embedded_obj(schema, embedded_obj):
    obj_info = {}
    obj_title = embedded_obj.get('title')
    obj_props = _get_submittable_props(schema, get_properties(embedded_obj))
    obj_props = _annotate_submittable_props(schema, obj_props)
    if obj_props:
        obj_info[sc.PROPERTIES] = obj_props
        if obj_title:
            obj_info['title'] = obj_title
    return obj_info


def _get_submittable_props(schema, props):
    """
    Use appproject provided info on properties and properties with certain attributes
    to exclude from submittable properties.  Excluded properties is a list of prop names
    and excluded attibutes is a dictionary of attribute names and values
    """
    excluded_props = app_project().get_properties_for_exclusion()
    exclude_attrs = app_project().get_attributes_for_exclusion()

    submittable_props = {}

    for propname, propinfo in props.items():
        emb_obj = None
        # determine if prop should be submittable
        if propname in excluded_props:  # explicity excluded by name
            continue
        elif _has_property_attr_with_val(propinfo, exclude_attrs):
            continue
        elif is_array_schema(propinfo):  # need to check the attributes of the items
            list_item = get_items(propinfo)
            if is_object_schema(list_item):  # very rare case of list of embedded objects
                emb_obj = _build_embedded_obj(schema, list_item)
                if not emb_obj:
                    continue
                else:
                    propinfo[sc.ITEMS] = emb_obj
                    submittable_props[propname] = propinfo
                    emb_obj = None
            elif _has_property_attr_with_val(list_item, exclude_attrs):
                continue
            else:
                submittable_props[propname] = propinfo
        elif is_object_schema(propinfo):  # infrequent case of embedded object
            emb_obj = _build_embedded_obj(schema, propinfo)
            if emb_obj:
                submittable_props[propname] = emb_obj
        else:
            submittable_props[propname] = propinfo
    return submittable_props


def _get_submittable_schema(schema):
    """
    helper function to take a schema and determine whether it is submittable
    and if so parse info so only submittable fields are included along with hints
    and doc on those fields
    """
    schema_id = schema.get('$id')
    schema_props = get_properties(schema)
    submittable_schema = {}
    if not schema_props:
        return {}

    if not _is_submittable_schema(schema_id, schema):
        return {}

    submittable_props = _get_submittable_props(schema, schema_props)

    if not submittable_props:
        return {}

    submittable_props = _annotate_submittable_props(schema, submittable_props)

    if schema_id:
        submittable_schema['$id'] = schema_id
    submittable_schema['title'] = schema.get('title')
    submittable_schema[sc.PROPERTIES] = submittable_props

    return submittable_schema


@view_config(route_name='submittable', request_method='GET',
             decorator=etag_app_version_effective_principals)
@debug_log
def submittable(context, request):
    schema2chk = schema(context, request)
    submittable_schema = _get_submittable_schema(schema2chk)
    if not submittable_schema:
        raise HTTPNotFound(f'The schema you requested with {request.url} is not submittable or has no submittable fields')
    return submittable_schema


@view_config(route_name='submittables', request_method='GET',
             decorator=etag_app_version_effective_principals)
@debug_log
def submittables(context, request):
    submittable_schemas = {}
    all_schemas = schemas(context, request)
    for name, schema in all_schemas.items():
        submittable_schema = _get_submittable_schema(schema)
        if submittable_schema:
            submittable_schemas[name] = submittable_schema
    if not submittable_schemas:
        raise HTTPNotFound("No submittable schemas found")
    return submittable_schemas



from unittest import mock
import pytest
import re
import uuid
import boto3

from dcicutils.misc_utils import filtered_warnings
from pyramid.threadlocal import manager
from sqlalchemy import func
from sqlalchemy.exc import IntegrityError
from ..interfaces import DBSESSION, STORAGE
from ..storage import (
    POSTGRES_COMPATIBLE_MAJOR_VERSIONS,
    Blob,
    CurrentPropertySheet,
    Key,
    Link,
    PickStorage,
    PropertySheet,
    RDBStorage,
    register_storage,
    Resource,
    S3BlobStorage,
)
from moto import mock_s3

pytestmark = pytest.mark.storage


def test_postgres_version(session):
    """ Tests that the local postgres is running one of the compatible versions """
    (version_info,) = session.query(func.version()).one()
    print("version_info=", version_info)
    assert isinstance(version_info, str)
    assert re.match("PostgreSQL (%s)([.][0-9]+)? " % '|'.join(POSTGRES_COMPATIBLE_MAJOR_VERSIONS), version_info)


def test_storage_creation(session):
    assert session.query(PropertySheet).count() == 0
    assert session.query(CurrentPropertySheet).count() == 0
    assert session.query(Blob).count() == 0
    assert session.query(Key).count() == 0
    assert session.query(Link).count() == 0


def test_transaction_record_rollback(session):
    """ Tests that committing and rolling back an invalid transactions works as expected """
    rid = uuid.uuid4()
    sp1 = session.begin_nested()
    resource = Resource('test_item', {'': {}}, rid=rid)
    session.add(resource)
    sp1.commit()

    # test rollback
    sp2 = session.begin_nested()
    resource = Resource('test_item', {'': {}}, rid=rid)
    session.add(resource)
    with pytest.raises(Exception):
        sp2.commit()
    sp2.rollback()
    resource = Resource('test_item', {'': {}})
    session.add(resource)


def test_current_propsheet(session):
    name = 'testdata'
    props1 = {'foo': 'bar'}
    resource = Resource('test_item', {name: props1})
    session.add(resource)
    session.flush()
    resource = session.query(Resource).one()
    assert resource.rid
    assert resource[name] == props1
    propsheet = session.query(PropertySheet).one()
    assert propsheet.sid
    assert propsheet.rid == resource.rid
    current = session.query(CurrentPropertySheet).one()
    assert current.sid == propsheet.sid
    assert current.rid == resource.rid
    # tid is removed
    assert not hasattr(current, 'tid')


def test_current_propsheet_update(session):
    name = 'testdata'
    props1 = {'foo': 'bar'}
    resource = Resource('test_item', {name: props1})
    session.add(resource)
    session.flush()
    resource = session.query(Resource).one()
    props2 = {'foo': 'baz'}
    resource[name] = props2
    session.flush()
    resource = session.query(Resource).one()
    session.flush()
    assert resource[name] == props2
    assert session.query(PropertySheet).count() == 2
    assert [propsheet.properties for propsheet in resource.data[name].history] == [props1, props2]
    current = session.query(CurrentPropertySheet).one()
    assert current.sid
    # tid is removed
    assert not hasattr(current, 'tid')


def test_get_by_json(session):
    name = 'testdata'
    props1 = {'foo': 'bar'}
    resource = Resource('test_item', {name: props1})
    session.add(resource)
    session.flush()
    resource = session.query(Resource).one()
    props2 = {'foo': 'baz'}
    resource[name] = props2
    session.flush()
    resource = session.query(Resource).one()
    session.flush()
    query = (session.query(CurrentPropertySheet)
             # Rewrittent to use two separate joins per SQLAlchemy 2.0 requirements. -kmp 10-Apr-2023
             .join(CurrentPropertySheet.propsheet)
             .join(CurrentPropertySheet.resource)
             .filter(PropertySheet.properties['foo'].astext == 'baz')
             )
    data = query.one()
    assert data.propsheet.properties == props2


def test_purge_uuid(session, storage):
    """ Tests full purge of metadata (including revision history). """
    name = 'testdata'
    props1 = {'foo': 'bar'}
    resource = Resource('test_item', {name: props1})
    session.add(resource)
    session.flush()
    resource = session.query(Resource).one()
    check = storage.get_by_uuid(str(resource.rid))
    assert check[name] == props1
    # add a key
    testname = 'foo'
    key = Key(rid=resource.rid, name=testname, value=props1[testname])
    session.add(key)
    session.flush()
    assert session.query(Key).count() == 1

    propsheet = session.query(PropertySheet).one()
    assert propsheet.sid
    assert propsheet.rid == resource.rid
    current = session.query(CurrentPropertySheet).one()
    assert current.sid == propsheet.sid
    assert current.rid == resource.rid

    assert len(storage.revision_history(uuid=str(resource.rid))) > 0
    storage.purge_uuid(str(resource.rid))
    check_post = storage.get_by_uuid(str(resource.rid))
    assert not check_post
    assert session.query(Key).count() == 0
    assert session.query(PropertySheet).count() == 0
    assert session.query(CurrentPropertySheet).count() == 0
    assert storage.revision_history(uuid=str(resource.rid)) == []


def test_delete_compound(session, storage):
    name = 'testdata'
    props1 = {'foo': 'bar'}
    resource = Resource('test_item', {name: props1})
    session.add(resource)
    session.flush()
    resource = session.query(Resource).one()
    check = storage.get_by_uuid(str(resource.rid))
    assert check[name] == props1
    # add a key
    testname = 'foo'
    key = Key(rid=resource.rid, name=testname, value=props1[testname])
    session.add(key)
    session.flush()
    assert session.query(Key).count() == 1

    props2 = {'foo': 'baz'}
    resource[name] = props2
    session.flush()
    resource = session.query(Resource).one()
    session.flush()
    assert resource[name] == props2
    assert session.query(PropertySheet).count() == 2
    assert [propsheet.properties for propsheet in resource.data[name].history] == [props1, props2]
    current = session.query(CurrentPropertySheet).one()
    assert current.sid

    assert len(storage.revision_history(uuid=str(resource.rid))) > 0
    storage.purge_uuid(str(resource.rid))
    check_post = storage.get_by_uuid(str(resource.rid))
    assert not check_post
    assert session.query(Key).count() == 0
    assert session.query(PropertySheet).count() == 0
    assert session.query(CurrentPropertySheet).count() == 0
    assert storage.revision_history(uuid=str(resource.rid)) == []


def test_keys(session):
    name = 'testdata'
    props1 = {'foo': 'bar'}
    resource = Resource('test_item', {name: props1})
    session.add(resource)
    session.flush()
    resource = session.query(Resource).one()

    testname = 'foo'
    key = Key(rid=resource.rid, name=testname, value=props1[testname])
    session.add(key)
    session.flush()
    assert session.query(Key).count() == 1
    othertest = 'foofoo'
    othervalue = 'barbar'
    key2 = Key(rid=resource.rid, name=othertest, value=othervalue)
    session.add(key2)
    session.flush()
    assert session.query(Key).count() == 2
    props2 = {'foofoo': 'barbar'}
    resource2 = Resource('test_item', {name: props2})
    session.add(resource2)
    session.flush()
    key3 = Key(rid=resource2.rid, name=testname, value=props1[testname])
    session.add(key3)
    # try to insert a duplicate unique key, previously threw FlushError
    with pytest.raises(IntegrityError):  # in newer sqlalchemy versions IntegrityError is more accurate
        session.flush()


def test_get_sids_by_uuids(session, storage):
    props1 = {'foo': 'bar'}
    resource = Resource('test_item', {'': props1})
    session.add(resource)
    session.flush()
    resource = session.query(Resource).one()
    sids = storage.get_sids_by_uuids([str(resource.rid)])
    assert set(sids) == {str(resource.rid)}


@pytest.mark.parametrize(
    's3_encrypt_key_id,kms_args_expected',
    [(None, False), ("", False), (str(uuid.uuid4()), True)],
)
def test_S3BlobStorage(s3_encrypt_key_id, kms_args_expected):
    # NOTE: I have separated the call to _test_S3BlobStorage into a separate function so I can wrap it with
    #       @mock_s3 after establishing this context manager to suppress a warning. (It could have been an
    #       internal function, but it shows up better in the patch diffs if I do it this way, and will be
    #       easier to simplify later. The warning is due to a call made by moto 1.3.x that uses deprecated
    #       support complained about in responses 0.17.0. Hopefully if we ever get higher than that version
    #       of moto, we can reconsider this. However, moto 2.0 requires some change in configuration that we'd
    #       have to take time to learn about. -kmp 5-Feb-2022
    with filtered_warnings('ignore', category=DeprecationWarning):
        # The warning being suppressed (which comes from moto 1.3.x) looks like:
        #   ...env/lib/python3.6/site-packages/responses/__init__.py:484:
        #   DeprecationWarning: stream argument is deprecated. Use stream parameter in request directly
        # HOPEFULLY that's the only deprecation warning that would come from this test, which is why it
        # would be good to remove these warnings when we are able. -kmp 5-Feb-2022
        _test_S3BlobStorage(s3_encrypt_key_id, kms_args_expected)


@mock_s3
def _test_S3BlobStorage(s3_encrypt_key_id, kms_args_expected):

    blob_bucket = 'encoded-4dn-blobs'  # note that this bucket exists but is mocked out here
    conn = boto3.resource('s3', region_name='us-east-1')
    conn.create_bucket(Bucket=blob_bucket)

    storage = S3BlobStorage(blob_bucket, kms_key_id=s3_encrypt_key_id)
    assert storage.bucket == blob_bucket
    if s3_encrypt_key_id:
        assert storage.kms_key_id == s3_encrypt_key_id

    download_meta = {'download': 'test.txt'}
    with mock.patch.object(
        storage.s3, 'put_object', side_effect=storage.s3.put_object
    ) as mocked_s3_put_object:  # To obtain calls while retaining function
        storage.store_blob('data', download_meta)
        assert download_meta['bucket'] == blob_bucket
        assert 'key' in download_meta
        mocked_s3_put_object.assert_called_once()
        call_kwargs = mocked_s3_put_object.call_args.kwargs
        if kms_args_expected:
            assert call_kwargs.get("ServerSideEncryption") == "aws:kms"
            assert call_kwargs.get("SSEKMSKeyId") == s3_encrypt_key_id
        else:
            assert "ServerSideEncryption" not in call_kwargs
            assert "SSEKMSKeyId" not in call_kwargs

    data = storage.get_blob(download_meta)
    assert data == 'data'

    url = storage.get_blob_url(download_meta)
    assert url
    assert blob_bucket in url
    assert 'Signature' in url


def test_S3BlobStorage_get_blob_url_for_non_s3_file():
    # NOTE: See test_S3BlobStorage for note explaining this filtering of warning and when/how it can go away.
    #       -kmp 5-Feb-2022
    with filtered_warnings('ignore', category=DeprecationWarning):
        # The warning being suppressed (which comes from moto 1.3.x) looks like:
        #   ...env/lib/python3.6/site-packages/responses/__init__.py:484:
        #   DeprecationWarning: stream argument is deprecated. Use stream parameter in request directly
        _test_S3BlobStorage_get_blob_url_for_non_s3_file()


@mock_s3
def _test_S3BlobStorage_get_blob_url_for_non_s3_file():

    blob_bucket = 'encoded-4dn-blobs'
    storage = S3BlobStorage(blob_bucket)
    assert storage.bucket == blob_bucket
    download_meta = {'blob_id': 'blob_id'}
    url = storage.get_blob_url(download_meta)
    assert url


def test_pick_storage(registry, dummy_request):
    # use a dummy value for ElasticSearchStorage
    storage = PickStorage(RDBStorage(registry[DBSESSION]), 'dummy_es', registry)
    assert isinstance(storage.write, RDBStorage)
    assert storage.read == 'dummy_es'
    # test storage selection logic
    assert storage.storage('database') is storage.write
    assert storage.storage('elasticsearch') == 'dummy_es'
    with pytest.raises(Exception) as exec_info:
        storage.storage('not_a_db')
    assert 'Invalid forced datastore not_a_db' in str(exec_info.value)
    assert storage.storage() is storage.write

    dummy_request.datastore = 'elasticsearch'
    manager.push({'request': dummy_request, 'registry': registry})
    assert storage.storage() == 'dummy_es'
    manager.pop()


def test_register_storage(registry):
    # test storage.register_storage, used to configure registry[STORAGE]
    storage = PickStorage('dummy_db', 'dummy_es', registry)
    # store previous storage and use a dummy one for testing
    prev_storage = registry[STORAGE]
    registry[STORAGE] = storage
    # expect existing values to be used
    register_storage(registry)
    assert registry[STORAGE].write == 'dummy_db'
    assert registry[STORAGE].read == 'dummy_es'
    # expect overrides to be used
    register_storage(registry, write_override='override_db',
                     read_override='override_es')
    assert registry[STORAGE].write == 'override_db'
    assert registry[STORAGE].read == 'override_es'
    # reset storage
    registry[STORAGE] = prev_storage



import argparse
import logging
import structlog
import transaction

from dcicutils.env_utils import is_stg_or_prd_env
from dcicutils.lang_utils import disjoined_list
from pyramid.paster import get_app
from snovault import DBSESSION
from sqlalchemy import text as psql_text
from ..storage import Base
from ..elasticsearch.create_mapping import run as run_create_mapping
from sqlalchemy import MetaData
from typing import Optional, List
from zope.sqlalchemy import mark_changed
from .. import configure_dbsession


log = structlog.getLogger(__name__)


EPILOG = __doc__


def clear_db_tables(app):
    """
    Given a pyramids app that has a configured DB session, will clear the
    contents of all DB tables

    Args:
        app: Pyramid application

    Returns:
        bool: True if successful, False if error encountered
    """
    success = False
    session = app.registry[DBSESSION]
    meta = MetaData(bind=session.connection())
    meta.reflect()
    connection = session.connection().connect()
    try:
        # truncate tables by only deleting contents (sqlalchemy 1.4+ compliant)
        table_names = ','.join(table.name for table in reversed(Base.metadata.sorted_tables))
        connection.execute(psql_text('SET statement_timeout = 300000;'))  # give 5 mins for DB clear
        connection.execute(psql_text(f'TRUNCATE {table_names} RESTART IDENTITY;'))
    except Exception as e:
        log.error(f"clear_db_es_contents: Error on DB drop_all/create_all. {type(e)}: {e}")
        transaction.abort()
    else:
        # commit all changes to DB
        session.flush()
        mark_changed(session())
        transaction.commit()
        success = True
    return success


SKIPPING_CLEAR_ATTEMPT = 'Skipping the attempt to clear DB.'


def run_clear_db_es(app, only_envs: Optional[List[str]] = None, skip_es: bool = False,
                    allow_prod: bool = False) -> bool:
    """
    This function actually clears DB/ES. Takes a Pyramid app as well as two flags. _Use with care!_

    For safety, this function will return without side-effect if ...
    - The current environment is any production system.
    - The current environment is not a member of the `only_envs` argument (list).

    If `arg_skip_es` (default False) is True, this function will return after DB clear
    and before running create_mapping.

    Args:
        app: Pyramid application
        only_envs (list): a list of env names that are the only envs where this action will run
        skip_es (bool): if True, do not run create_mapping after DB clear
        allow_prod (bool): if True, allows running on envs that are set to the staging or prod
                           env in the GLOBAL_ENV_BUCKET (main.ecosystem)

    Returns:
        bool: True if DB was cleared (regardless of ES)
    """
    current_env = app.registry.settings.get('env.name', 'local')

    if is_stg_or_prd_env(current_env) and not allow_prod:
        log.error(f"clear_db_es_contents: This action cannot be performed on env {current_env}"
                  f" because it is a production-class (stg or prd) environment."
                  f" {SKIPPING_CLEAR_ATTEMPT}")
        return False

    if only_envs and current_env not in only_envs:
        log.error(f"clear_db_es_contents: The current environment, {current_env}, is not {disjoined_list(only_envs)}."
                  f" {SKIPPING_CLEAR_ATTEMPT}")
        return False

    log.info('clear_db_es_contents: Clearing DB tables...')
    db_success = clear_db_tables(app)
    if not db_success:
        log.error("clear_db_es_contents: Clearing DB tables failed!"
                  " Such failures may happen, for example, when there are external DB connections."
                  " You might want to try running clear_db_es_contents again.")
        return False
    log.info("clear_db_es_contents: Successfully cleared DB.")

    # create mapping after clear DB to remove ES contents
    if not skip_es:
        log.info("clear_db_es_contents: Clearing ES with create_mapping...")
        run_create_mapping(app, purge_queue=True)
        log.info("clear_db_es_contents: Successfully cleared ES.")

    log.info("clear_db_es_contents: All done.")
    return True


def main(simulated_args=None):
    logging.basicConfig()
    # Loading app will have configured from config file. Reconfigure here:
    logging.getLogger('encoded').setLevel(logging.DEBUG)

    parser = argparse.ArgumentParser(  # noqa - PyCharm wrongly thinks the formatter_class is specified wrong here.
        description='Clear DB and ES Contents', epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument('config_uri', help='path to configfile')
    parser.add_argument('--app-name', help='Pyramid app name in configfile')
    parser.add_argument('--only-if-env', '--only-if-envs', dest='only_envs', default=None,
                        help=("A comma-separated list of envs where this action is allowed to run."
                              " If omitted, any env is OK to run."))
    parser.add_argument("--confirm", action="store_true", dest="confirm", default=None,
                        help="Specify --confirm to require interactive confirmation.")
    parser.add_argument("--no-confirm", action="store_false", dest="confirm", default=None,
                        help="Specify --no-confirm to suppress interactive confirmation.")
    parser.add_argument('--skip-es', action='store_true', default=False,
                        help='If set, do not run create_mapping after DB drop')
    parser.add_argument('--allow-prod', action='store_true', default=False,
                        help='DANGER: If set, will allow running this command on an env that is staging or prod')
    args = parser.parse_args(simulated_args)

    confirm = args.confirm
    app_name = args.app_name
    config_uri = args.config_uri
    only_envs = args.only_envs
    skip_es = args.skip_es
    allow_prod = args.allow_prod

    if confirm is None:
        confirm = not only_envs  # If only_envs is supplied, we have better protection so don't need to confirm

    # get the pyramids app
    app = get_app(config_uri, app_name)

    # create db schema
    configure_dbsession(app)

    only_envs = [x for x in (only_envs or "").split(',') if x]

    if confirm:
        env_to_confirm = app.registry.settings.get('env.name', 'local')
        env_confirmation = input(f'This will completely clear DB contents for environment {env_to_confirm}.\n'
                                 f' Type the env name to confirm: ')
        if env_confirmation != env_to_confirm:
            print(f"NOT confirmed. {SKIPPING_CLEAR_ATTEMPT}")
            return

    # actually run. split this out for easy testing
    run_clear_db_es(app=app, only_envs=only_envs, skip_es=skip_es, allow_prod=allow_prod)


if __name__ == '__main__':
    main()



"""
Cached views used when model was pulled from elasticsearch.
In some cases, use these views to render object and embedded content.
Leverages interfaces.ICachedItem as a provided interface for resources
from Elasticsearch. See esstorage.CachedModel and the following docs:
https://zopeinterface.readthedocs.io/en/stable/api.html#zope.interface.declarations.alsoProvides
"""

from dcicutils.misc_utils import ignored
from pyramid.httpexceptions import HTTPForbidden
from pyramid.view import view_config

from .interfaces import ICachedItem
from ..resource_views import (
    item_view_embedded,
    item_view_object,
    item_view_page,
    item_view_expand
)
from ..indexing_views import item_index_data
from ..util import debug_log, filter_embedded


def includeme(config):
    config.scan(__name__)


# As discussed in code review, we THINK this was unused (is redefined by the subsequent definition of same name).
#
# @view_config(context=ICachedItem, request_method='GET',
#              name='embedded')
# @debug_log
# def cached_view_embedded(context, request):  # This name duplicates the next definition.
#     source = context.model.source
#     allowed = set(source['principals_allowed']['view'])
#     if allowed.isdisjoint(request.effective_principals):
#         raise HTTPForbidden()
#     return filter_embedded(source['embedded'], request.effective_principals)


@view_config(context=ICachedItem, permission='view', request_method='GET',
             name='embedded')
@debug_log
def cached_view_embedded(context, request):  # This name duplicates the previous definition.
    """
    Use the 'embedded' view that is stored the ElasticSearch unless we
    are using an Item with 'elasticsearch' properties_datastore, in which case
    we may have to generate the view dynamically if indexing or it doesn't yet
    exist
    """
    source = context.model.source
    # generate view if this item uses ES as primary datastore or indexing
    if (context.properties_datastore == 'elasticsearch'
            and ('embedded' not in source or request._indexing_view is True)):
        embedded = item_view_embedded(context, request)
    else:
        embedded = source['embedded']

    # permission checking on embedded objects
    allowed = set(embedded['principals_allowed']['view'])
    if allowed.isdisjoint(request.effective_principals):
        raise HTTPForbidden()
    return filter_embedded(embedded, request.effective_principals)


@view_config(context=ICachedItem, permission='view', request_method='GET',
             name='object')
@debug_log
def cached_view_object(context, request):
    """
    Use the 'object' view that is stored the ElasticSearch unless we
    are using an Item with 'elasticsearch' used_datastore, in which case we
    may have to generate the view dynamically if indexing or it doesn't yet
    exist
    """
    source = context.model.source
    # generate view if this item uses ES as primary datastore or indexing
    if (context.properties_datastore == 'elasticsearch'
            and ('object' not in source or request._indexing_view is True)):
        object = item_view_object(context, request)
    else:
        object = source['object']

    # permission checking. Is this redundant with permission set on view?
    allowed = set(object['principals_allowed']['view'])
    if allowed.isdisjoint(request.effective_principals):
        raise HTTPForbidden()
    return object


@view_config(context=ICachedItem, permission='view_raw', request_method='GET',
             name='raw')
def cached_view_raw(context, request):
    """
    Must generate this view because it's not stored in ES source. Add 'uuid'
    to the item properties to be consistent with the regular 'raw' view
    """
    ignored(request)
    source = context.model.source
    props = source['properties']
    # add uuid to raw view
    props['uuid'] = source['uuid']
    return props


@view_config(context=ICachedItem, permission='view', request_method='GET',
             name='page')
def cached_view_page(context, request):
    """
    Must generate this view because it's not stored in ES source
    """
    return item_view_page(context, request)


@view_config(context=ICachedItem, permission='expand', request_method='GET',
             name='expand')
def cached_view_expand(context, request):
    """
    Must generate this view because it's not stored in ES source
    """
    return item_view_expand(context, request)


@view_config(context=ICachedItem, permission='index', request_method='GET',
             name='index-data')
def cached_index_data(context, request):
    """
    Must generate this view because it's not stored in ES source
    """
    return item_index_data(context, request)



# Web browsers send an Accept request header for initial (e.g. non-AJAX) page requests
# which should contain 'text/html'
MIME_TYPE_HTML = 'text/html'
MIME_TYPE_JSON = 'application/json'
MIME_TYPE_LD_JSON = 'application/ld+json'



import argparse
import logging
import structlog
# import transaction

from dcicutils.env_utils import is_stg_or_prd_env
from dcicutils.lang_utils import disjoined_list
from dcicutils.misc_utils import PRINT, get_error_message
from pyramid.paster import get_app
# from snovault import DBSESSION
# from snovault.storage import Base
# from snovault.elasticsearch.create_mapping import run as run_create_mapping
# from sqlalchemy import MetaData
from typing import Optional, List
# from zope.sqlalchemy import mark_changed
from .. import configure_dbsession
from ..sqlalchemy_tools import PyramidAppManager
from ..project_app import app_project


logger = structlog.getLogger(__name__)


EPILOG = __doc__


def list_db_tables(app):
    """
    Given a pyramids app that has a configured DB session, will list the contents of all DB tables

    Args:
        app: Pyramid application

    Returns:
        bool: True if successful, False if error encountered
    """

    app_manager = PyramidAppManager(app)

    with app_manager.connection() as connection:
        for table_name in app_manager.ordered_table_names:
            n = connection.execute(f"SELECT COUNT(*) FROM {table_name};").one()
            print(f" Table {table_name}: {n}")


SKIPPING_LIST_ATTEMPT = 'Skipping the attempt to list DB.'


def run_list_db_tables(app, only_envs: Optional[List[str]] = None, skip_es: bool = False,
                    allow_prod: bool = False) -> bool:
    """
    This function lists information from the DB.

    For safety, this function will return without side-effect if ...
    - The current environment is any production system (and allow_prod is not given).
    - The current environment is not a member of the `only_envs` argument (list).

    Args:
        app: Pyramid application
        only_envs (list): a list of env names that are the only envs where this action will run
        allow_prod (bool): if True, allows running on envs that are set to the staging or prod
                           env in the GLOBAL_ENV_BUCKET (main.ecosystem)

    Returns:
        bool: True if DB was listed.
    """
    current_env = app.registry.settings.get('env.name', 'local')

    if is_stg_or_prd_env(current_env) and not allow_prod:
        logger.error(f"list-db-tables: This action cannot be performed on env {current_env}"
                     f" because it is a production-class (stg or prd) environment."
                     f" {SKIPPING_LIST_ATTEMPT}")
        return False

    if only_envs and current_env not in only_envs:
        logger.error(f"list-db-tables: The current environment, {current_env}, is not {disjoined_list(only_envs)}."
                     f" {SKIPPING_LIST_ATTEMPT}")
        return False

    logger.info('list-db-tables: Listing DB tables...')
    try:
        list_db_tables(app)
    except Exception as e:
        logger.info(f"list-db-tables failed. {get_error_message(e)}")
        return False
    logger.info("list-db-tables succeeded.")
    return True


def main(simulated_args=None):
    parser = argparse.ArgumentParser(  # noqa - PyCharm wrongly thinks the formatter_class is specified wrong here.
        description='List DB Contents', epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument('--config_uri', help='path to configfile', default='development.ini')
    parser.add_argument('--app-name', help='Pyramid app name in configfile', default='app')
    parser.add_argument('--only-if-env', '--only-if-envs', dest='only_envs', default=None,
                        help=("A comma-separated list of envs where this action is allowed to run."
                              " If omitted, any env is OK to run."))
    parser.add_argument("--confirm", action="store_true", dest="confirm", default=None,
                        help="Specify --confirm to require interactive confirmation.")
    parser.add_argument("--no-confirm", action="store_false", dest="confirm", default=None,
                        help="Specify --no-confirm to suppress interactive confirmation.")
    parser.add_argument('--allow-prod', action='store_true', default=False,
                        help='DANGER: If set, will allow running this command on an env that is staging or prod')
    parser.add_argument('--log', action='store_true', default=False,
                        help='Set loglevel to DEBUG. Otherwise it will be ERROR.')
    args = parser.parse_args(simulated_args)

    confirm = args.confirm
    app_name = args.app_name
    config_uri = args.config_uri
    only_envs = args.only_envs
    allow_prod = args.allow_prod
    log = args.log

    logging.basicConfig()
    #project = app_project(initialize=True)
    project = app_project()
    # Loading app will have configured from config file. Reconfigure here:
    if log:
        logging.getLogger(project.NAME).setLevel(logging.DEBUG)
    else:
        logging.getLogger(project.NAME).setLevel(logging.ERROR)

    if confirm is None:
        confirm = False   # not only_envs  # If only_envs is supplied, we have better protection so don't need to confirm

    # get the pyramids app
    app = get_app(config_uri, app_name)

    # create db schema
    configure_dbsession(app)

    only_envs = [x for x in (only_envs or "").split(',') if x]

    if confirm:
        env_to_confirm = app.registry.settings.get('env.name', 'local')
        env_confirmation = input(f'This will list DB contents for environment {env_to_confirm}.\n'
                                 f' Type the env name to confirm: ')
        if env_confirmation != env_to_confirm:
            PRINT(f"NOT confirmed. {SKIPPING_LIST_ATTEMPT}")
            return

    # actually run. split this out for easy testing
    run_list_db_tables(app=app, only_envs=only_envs, allow_prod=allow_prod)


if __name__ == '__main__':
    main()



import argparse
import logging
import structlog
import json
import os
import boto3
from pyramid.paster import get_app
from webtest import AppError
from dcicutils.misc_utils import TestApp
from dcicutils.beanstalk_utils import get_beanstalk_real_url
from dcicutils.cloudformation_utils import get_ecs_real_url
from dcicutils.secrets_utils import assume_identity

log = structlog.getLogger(__name__)
EPILOG = __doc__


# This should be imported/overridden by downstream application
ADMIN_EMAIL = 'snovault.platform@gmail.com'


def get_existing_key_ids(testapp, user_uuid, key_desc):
    """
    Search for an access key with given description and user uuid.
    If successful return list of @id values of all found access keys.
    This is expected to fail if the current server is not up.
    Logs information on errors.

    Args:
        testapp (webtest.TestApp): current TestApp
        user_uuid (str): uuid of the user used to generate the key
        key_desc (str): description of the access key to find

    Return:
        list: of str access keys ids
    """
    try:
        search_res = testapp.get('/search/?type=AccessKey&description=%s&user.uuid=%s'
                                 % (key_desc, user_uuid)).json
    except Exception as exc:
        log.error('load_access_keys: search failed for access key with desc'
                  ' %s. Exception: %s' % (key_desc, exc))
        return []
    if len(search_res['@graph']) > 1:
        log.warning('load_access_keys: %s access keys found with '
                    'description %s and user.uuid %s.'
                    % (len(search_res['@graph']), key_desc, user_uuid))
    return [res['@id'] for res in search_res['@graph']]


def generate_access_key(testapp, env, user_uuid, description):
    """
    Generate an access for given user on given environment.

    Args:
        testapp (webtest.TestApp): current TestApp
        env (str): application environment used to find server
        user_uuid (str): uuid of the user to gener
        description (str): description to add to access key

    Returns:
        dict: access key contents with server
    """
    try:
        server = get_ecs_real_url(env)  # try to grab from Cfn, if we are ECS env
    except Exception:
        server = get_beanstalk_real_url(env)
    if not server:
        server = get_beanstalk_real_url(env)
    access_key_req = {'user': user_uuid, 'description': description}
    res = testapp.post_json('/access_key', access_key_req).json
    return {'secret': res['secret_access_key'],
            'key': res['access_key_id'],
            'server': server}


def main():
    """
    Function to create and load access keys for multiple users to the system s3
    bucket. The description of the key is set to the s3 object name. Before
    creating the keys, will attempt to delete pre-existing keys with the given
    descriptions. This command is not meant to be run locally.

    Provide required `config_uri` and `--app-name` to the command.
    Example usage:
    `bin/load_access_keys production.ini --app-name app`
    """
    logging.basicConfig()
    # Loading app will have configured from config file. Reconfigure here:
    logging.getLogger('encoded').setLevel(logging.INFO)

    parser = argparse.ArgumentParser(  # noqa - PyCharm wrongly thinks the formatter_class is specified wrong here.
        description="Load Access Keys", epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument('config_uri', help='path to configfile')
    parser.add_argument('--app-name', help='Pyramid app name in configfile')
    parser.add_argument('--secret-name', help='name of application identity stored in secrets manager within which'
                                              'to locate S3_ENCRYPT_KEY, for example: dev/beanstalk/cgap-dev')
    args = parser.parse_args()

    app = get_app(args.config_uri, args.app_name)
    environ = {
        'HTTP_ACCEPT': 'application/json',
        'REMOTE_USER': 'TEST',
    }
    testapp = TestApp(app, environ)

    env = app.registry.settings.get('env.name')
    if not env:
        raise RuntimeError('load_access_keys: cannot find env.name in settings')

    # Resolve secret from environment if one is not specified
    encrypt_key = None
    if args.secret_name is not None:
        identity = assume_identity()  # automatically detects GLOBAL_APPLICATION_CONFIGURATION
        encrypt_key = identity.get('S3_ENCRYPT_KEY', None)  # one of the secrets
    if not encrypt_key:
        encrypt_key = os.environ.get('S3_ENCRYPT_KEY')

    if not encrypt_key:
        raise RuntimeError('load_access_keys: must define S3_ENCRYPT_KEY in env or in GAC')

    # will need to use a dynamic region at some point (not just here)
    s3 = boto3.client('s3', region_name='us-east-1')
    s3_bucket = app.registry.settings['system_bucket']

    # we generate keys for the following accounts w/ corresponding descriptions
    to_generate = [(ADMIN_EMAIL, 'access_key_admin'),
                   ('tibanna.app@gmail.com', 'access_key_tibanna'),
                   ('foursight.app@gmail.com', 'access_key_foursight')]
    for email, key_name in to_generate:
        try:
            user_props = testapp.get('/users/%s?datastore=database' % email).follow().json
        except Exception as exc:
            log.error('load_access_keys: could not get user %s. Exception: %s' % (email, exc))
            continue

        key_ids = get_existing_key_ids(testapp,  user_props['uuid'], key_name)
        for key_id in key_ids:
            try:
                testapp.patch_json(key_id, {'status': 'deleted'})
            except AppError:
                log.error('load_access_keys: key_id: %s does not exist in database but exists in ES' % key_id)

        key = generate_access_key(testapp, env, user_props['uuid'], key_name)
        s3.put_object(Bucket=s3_bucket,
                      Key=key_name,
                      Body=json.dumps(key),
                      SSECustomerKey=encrypt_key,
                      SSECustomerAlgorithm='AES256')
        log.info('load_access_keys: successfully generated access key %s' % key_name)


if __name__ == "__main__":
    main()



from ..types.ingestion import IngestionSubmission
from .exceptions import UndefinedIngestionProcessorType


_INGESTION_UPLOADERS = {}


def ingestion_processor(processor_type):
    """
    @ingestion_uploader(<ingestion-type-name>) is a decorator that declares the upload handler for an ingestion type.
    """

    # Make sure the ingestion type specified for the decorated function is supported by
    # our IngestionSubmission type; this info comes from schemas/ingestion_submission.json.
    if not IngestionSubmission.supports_type(processor_type):
        raise UndefinedIngestionProcessorType(processor_type)

    def ingestion_type_decorator(fn):
        if processor_type in _INGESTION_UPLOADERS:
            raise RuntimeError(f"Ingestion type {processor_type} is already defined.")
        _INGESTION_UPLOADERS[processor_type] = fn
        return fn

    return ingestion_type_decorator


def get_ingestion_processor(processor_type):
    handler = _INGESTION_UPLOADERS.get(processor_type, None)
    if not handler:
        raise UndefinedIngestionProcessorType(processor_type)
    return handler



import structlog
from collections import OrderedDict
from elasticsearch import (
    TransportError,
    RequestError,
    ConnectionTimeout
)
from pyramid.httpexceptions import HTTPBadRequest
from snovault import TYPES
from snovault.util import crawl_schema, find_collection_subtypes
from snovault.embed import make_subrequest
from snovault.elasticsearch.indexer_utils import get_namespaced_index
from snovault.elasticsearch.create_mapping import determine_if_is_date_field
from dcicutils.misc_utils import deduplicate_list


log = structlog.getLogger(__name__)

# Constants

# from now on, use these constants when referring to elastic search
# query keywords when writing elastic search queries - Will 3-20-2020
QUERY = 'query'
FILTER = 'filter'
MUST = 'must'
MUST_NOT = 'must_not'
BOOL = 'bool'
MATCH = 'match'
SHOULD = 'should'
EXISTS = 'exists'
FIELD = 'field'
NESTED = 'nested'
PATH = 'path'
TERMS = 'terms'
RANGE = 'range'
STATS = 'stats'
AGGS = 'aggs'
REVERSE_NESTED = 'reverse_nested'
# just for book-keeping/readability but is 'unused' for now
# ie: it should be obvious when you are 'effectively' writing lucene
ELASTIC_SEARCH_QUERY_KEYWORDS = [
    QUERY, FILTER, MUST, MUST_NOT, BOOL, MATCH, SHOULD, EXISTS, FIELD, NESTED, PATH, TERMS, RANGE, AGGS, REVERSE_NESTED,
]


COMMON_EXCLUDED_URI_PARAMS = [
    # Difference of this and URL params should result in all fields/filters.
    'frame', 'format', 'limit', 'sort', 'from', 'field',
    'mode', 'redirected_from', 'datastore', 'referrer',
    'currentAction', 'additional_facet', 'debug'
]
MAX_FACET_COUNTS = 100
RAW_FIELD_AGGREGATIONS = [
    'stats', 'nested:stats', 'date_histogram', 'histogram', 'range', 'nested:range',
]


# Exception Classes


class SearchException(Exception):
    """ Base Search Exception - not meant to be used directly """
    def __init__(self, *, func, msg=None):
        if msg is None:
            msg = 'Exception occurred in search code at stage %s' % func
        super(SearchException, self).__init__(msg)
        self.func = func


class QueryConstructionException(SearchException):
    """
    Query construction exception - throw this if we should throw an exception in query building
    due to invalid query params
    """
    def __init__(self, *, query_type, func, msg=None):
        if msg is None:
            msg = 'Exception occurred during query building at query type %s in func %s' % (query_type, func)
        super(QueryConstructionException, self).__init__(func=func, msg=msg)
        self.query_type = query_type


# Functions


def search_log(*, log_handler, msg, error=True):
    """ Utility function intended to prepend SEARCH to all log messages. All log messages originating
        in search code should use this method to log.

        :param log_handler: log handler to use
        :param msg: msg to log
        :param error: whether or not to log to error log. Default True, otherwise goes to DEBUG
    """
    if error:
        log_handler.error('SEARCH: ' + msg)
    else:
        log_handler.debug('SEARCH: ' + msg)


def convert_search_to_dictionary(search):
    """ Converts the given search to a dictionary. Useful in mocking queries from dictionaries in testing.

    :param search: elasticsearch_dsl object to convert
    :return: query in dictionary form
    """
    return search.to_dict()


def find_nested_path(field, es_mapping):
    """
    Returns path to 'highest' level nested field, in other words the first field mapped with type=nested
    found by traversing the given field from the *top* level.

    This function relies on information about the structure of the es_mapping to extract
    the *path to the object who's mapping is nested*. The comments within this function try to explain
    that structure. This information is needed to construct nested queries (it is the path).
    It returns None if the given (sub)field is not a member of a type=nested mapped field.

    :param field: the *full path* to the field we are filtering/aggregating on.
                        For example: "experiments_in_set.biosample.biosource.individual.organism.name"
    :param es_mapping: dictionary representation of the es_mapping of the type we are searching on
    :return: path for nested query or None
    """
    location = es_mapping
    possible_nested_paths = []
    path = []
    for cursor in field.split('.'):
        if cursor == 'raw':  # if we get to this point we're definitely at a leaf and should stop
            break
        if cursor not in location:  # its possible we are at a sub-embedded object boundary. Check if it has properties.
            if 'properties' not in location:  # if it doesn't have properties, there's nowhere to go, so return None.
                return None
            location = location['properties']  # else move location forward, but do not add it to the PATH
        if cursor not in location:  # if we still don't see our 'level', we are not a nested field
            break   # accumulated path will be discarded (not added to possible_nested_paths)
        location = location[cursor]
        path.append(cursor)
        if location.get('type', None) == 'nested':  # this could be a path
            possible_nested_paths.append('.'.join(path))
    # the last path added is the closest in proximity to the field and thus is correct
    return possible_nested_paths[-1] if possible_nested_paths else None


def is_schema_field(field):
    """ Returns whether or not we should expect a schema to be found for the given field.
        Currently this only applies to validation_errors and aggregated_items.

    :param field: field name to check
    :return: False if this field doesn't a schema, True otherwise
    """
    # XXX: Consider doing this with regex? - Will 6/11/2020
    if field.startswith('validation_errors') or field.startswith('aggregated_items'):  # note that trailing '.' is gone
        return False
    return True


def extract_field_name(field):
    """ Pre-processes 'field' from URL query params. Solely handles converting 'type' to '@type' and
        discarding the not (!) qualifier.

    :param field: field name to process
    :return: correct field_name to search on
    """
    use_field = '@type' if field == 'type' else field  # 'type' field is really '@type' in the schema
    return use_field[:-1] if use_field.endswith('!') else use_field


def schema_for_field(field, request, doc_types, should_log=False):
    """
    Find the schema for the given field (in embedded '.' format). Uses
    ff_utils.crawl_schema from snovault and logs any cases where there is an
    error finding the field from the schema. Caches results based off of field
    and doc types used

    :param field: embedded field path, separated by '.'
    :param request: current Request object
    :param doc_types (list): @types for the search
    :param should_log (bool): logging will only occur if set to True

    :returns: Dictionary schema for the field, or None if not found
    """
    types = request.registry[TYPES]
    schemas = [types[dt].schema for dt in doc_types]
    field_schema = None

    # We cannot hash dict by list (of doc_types) so we convert to unique ordered string
    doc_type_string = ','.join(sorted(doc_types))  # use default sort

    # Check cache, initializing if necessary
    cache = getattr(request, '_field_schema_cache', {})
    cache_key = (field, doc_type_string)
    if cache is None:
        request._field_schema_cache = cache = {}
    if cache_key in cache:
        return cache[cache_key]

    # for 'validation_errors.*' and 'aggregated_items.*',
    # schema will never be found and logging isn't helpful
    if schemas and is_schema_field(field):
        use_field = extract_field_name(field)
        for schema in schemas:
            try:
                field_schema = crawl_schema(types, use_field, schema)
            except Exception as exc:  # cannot find schema. Log and Return None
                if should_log:
                    log.warning('Cannot find schema in search.py. Type: %s. Field: %s'
                            % (doc_types[0], field), field=field, error=str(exc))
            else:
                if field_schema is not None:
                    break

    # Cache result, even if not found, for this request.
    cache[cache_key] = field_schema

    return field_schema


def get_query_field(field, facet):
    """
    Converts a field from its generic field name to a more specific field name referencing its embedded nature

    :param field: generic field name, such as 'files.accession'
    :param facet: facet on this field
    :return: full path to field on ES mapping
    """
    if field == 'type':
        return 'embedded.@type.raw'
    elif not is_schema_field(field):
        return field + '.raw'
    elif facet.get('aggregation_type') in RAW_FIELD_AGGREGATIONS:
        return 'embedded.' + field
    else:
        return 'embedded.' + field + '.raw'


def find_index_by_doc_types(request, doc_types, ignore):
    """
    Find the correct index(es) to be search given a list of doc_types.
    The types in doc_types are the item class names, formatted like
    'Experiment HiC' and index names are the item types, formatted like
    'experiment_hi_c'.
    Ignore any collection names provided in the ignore param, an array.
    Formats output indexes as a string usable by elasticsearch
    """
    indexes = []
    for doc_type in doc_types:
        if doc_type in ignore:
            continue
        else:
            result = find_collection_subtypes(request.registry, doc_type)
            namespaced_results = map(lambda t: get_namespaced_index(request, t), result)
            indexes.extend(namespaced_results)
    indexes = deduplicate_list(indexes)
    index_string = ','.join(indexes)
    return index_string


def get_es_index(request, doc_types):
    """
    Gets ES index based on doc_type (one type per index)
    if doc_type is item, search all indexes by setting es_index to None
    If multiple, search all specified

    :param request: current request, to be passed
    :param doc_types: item types we are searching on
    :return: index name
    """
    if 'Item' in doc_types:
        return get_namespaced_index(request, '*')
    else:
        return find_index_by_doc_types(request, doc_types, ['Item'])


def get_es_mapping(es, es_index):
    """
    Get es mapping for given doc type (so we can handle type=nested)
    Note this is the mechanism by which we "enable" the ability to do nested searches
    ie: only enabled on single index searches. You could feasibly add more criteria.

    :param es: elasticsearch client
    :param es_index: index to get mapping from
    :return: the mapping for this item type or {} if we are not doing a single index search
    """
    if '*' in es_index or ',' in es_index:  # no type=nested searches can be done on * or multi-index
        return {}
    else:
        index = es.indices.get(es_index)
        return index[es_index]['mappings']['properties']


def get_search_fields(request, doc_types):
    """
    Returns set of columns that are being searched and highlights
    XXX: Unused
    """
    fields = {'uuid'}
    highlights = {}
    types = request.registry[TYPES]
    for doc_type in doc_types:
        type_info = types[doc_type]
        for value in type_info.schema.get('boost_values', ()):
            fields.add('embedded.' + value)
            highlights['embedded.' + value] = {}
    return fields, highlights


def is_linkto_or_object_array_root_field(field, types, doc_types):
    """
    Not used currently.

    :param field: field to check
    :param types: registry types
    :param doc_types: types we are searching on
    :return: infer whether or not this field is mapped with type=nested based on the schema alone
    """
    schema = types[doc_types[0]].schema
    field_root = field.split('.')[0]
    fr_schema = (schema and schema.get('properties', {}).get(field_root, None)) or None
    if fr_schema and fr_schema['type'] == 'array' and (fr_schema['items'].get('linkTo') is not None or fr_schema['items']['type'] == 'object'):
        return True
    return False


def execute_search(*, es, query, index, from_, size, session_id=None):
    """
    Execute the given Elasticsearch-dsl search. Raise HTTPBadRequest for any
    exceptions that arise.

    :param es: handle to es
    :param query: dictionary representing ES query
    :param index: index to search
    :param from_: search start index
    :param size: # of records to return
    :param session_id: session if we are paginating
    :returns: Dictionary search results
    """
    err_exp = None
    es_results = None
    try:
        # set timeout
        es_results = es.search(index=index, body=query, from_=from_, size=size, timeout='30s', preference=session_id)
    except ConnectionTimeout:
        err_exp = 'The search failed due to a timeout. Please try a different query.'
    except RequestError as exc:
        # try to get a specific error message. May fail in some cases
        try:
            err_detail = str(exc.info['error']['root_cause'][0]['reason'])
        except Exception:
            err_detail = str(exc)
        err_exp = 'The search failed due to a request error: ' + err_detail
    except TransportError as exc:
        # most general exception
        exc_status = getattr(exc, 'status_code')
        if exc_status == 'TIMEOUT':
            err_exp = 'The search failed due to a timeout. Please try a different query.'
        else:
            err_exp = 'The search failed due to a transport error: ' + str(exc)
    except Exception as exc:
        err_exp = str(exc)  # XXX: We should revisit if we think this is always safe... -Will 4-23-2020
    if err_exp:
        raise HTTPBadRequest(explanation=err_exp)
    return es_results


def make_search_subreq(request, path, **kwargs):
    subreq = make_subrequest(request, path, **kwargs)
    if hasattr(request, "_stats"):
        subreq._stats = request._stats
    subreq.registry = request.registry
    if hasattr(request, "context"):
        subreq.context = request.context
    else:
        subreq.context = None
    subreq.headers['Accept'] = 'application/json'
    return subreq


def is_numerical_field(field_schema):
    """ Helper method that checks field_schemas type and determines if it is a 'numerical' field. """
    return field_schema.get('type', 'n/a') in ("integer", "float", "number")


def is_array_of_numerical_field(field_schema):
    """ Helper method that checks if field schema is a numerical array field. """
    items = field_schema.get('items', None)
    if items:
        return is_numerical_field(items)
    return False


def is_date_field(field, field_schema):
    """ Helper method that determines if field_schema is  """
    return determine_if_is_date_field(field, field_schema)


def build_sort_dicts(requested_sorts, request, doc_types=[], text_search=None):
    '''
    `text_search` not applicable for compound filtersets atm.. afaik... maybe we handle it later.
    '''

    sort = OrderedDict()
    result_sort = OrderedDict()

    if len(doc_types) == 1:
        type_schema = request.registry[TYPES][doc_types[0]].schema
    else:
        type_schema = None

    def add_to_sort_dict(requested_sort):
        if requested_sort.startswith('-'):
            name = requested_sort[1:]
            order = 'desc'
        else:
            name = requested_sort
            order = 'asc'

        sort_schema = schema_for_field(name, request, doc_types)

        if sort_schema:
            sort_type = sort_schema.get('type')
        else:
            sort_type = 'string'

        # ES type != schema types
        if sort_type == 'integer':
            sort['embedded.' + name] = result_sort[name] = {
                'order': order,
                'unmapped_type': 'long',
                'missing': '_last'
            }
        elif sort_type == 'number':
            sort['embedded.' + name] = result_sort[name] = {
                'order': order,
                'unmapped_type': 'float',
                'missing': '_last'
            }
        elif sort_schema and determine_if_is_date_field(name, sort_schema):
            sort['embedded.' + name + '.raw'] = result_sort[name] = {
                'order': order,
                'unmapped_type': 'date',
                'missing': '_last'
            }
        else:
            # fallback case, applies to all string type:string fields
            sort['embedded.' + name + '.lower_case_sort'] = result_sort[name] = {
                'order': order,
                'unmapped_type': 'keyword',
                'missing': '_last'
            }


    # Prefer sort order specified in request, if any
    if requested_sorts:
        for rs in requested_sorts:
            add_to_sort_dict(rs)

    # Otherwise we use a default sort only when there's no text search to be ranked
    if not sort and (text_search == '*' or not text_search):
        # If searching for a single type, look for sort options in its schema
        if type_schema:
            if 'sort_by' in type_schema:
                for k, v in type_schema['sort_by'].items():
                    # Should always sort on raw field rather than analyzed field
                    # OR search on lower_case_sort for case insensitive results
                    sort['embedded.' + k + '.lower_case_sort'] = result_sort[k] = v
        # Default is most recent first, then alphabetical by label
        if not sort:
            sort['embedded.date_created.raw'] = result_sort['date_created'] = {
                'order': 'desc',
                'unmapped_type': 'keyword',
            }
            sort['embedded.label.raw'] = result_sort['label'] = {
                'order': 'asc',
                'missing': '_last',
                'unmapped_type': 'keyword',
            }

    return (sort, result_sort)



import json
import os
import pytest

from base64 import b64encode
from jsonschema import Draft202012Validator
from pyramid.compat import ascii_native_
from uuid import uuid4

from ..interfaces import TYPES
from ..util import mappings_use_nested

from .testing_views import (
    NESTED_EMBEDDING_CONTAINER_GUID,
    NESTED_OBJECT_LINK_TARGET_GUID_1,
    NESTED_OBJECT_LINK_TARGET_GUIDS,
)


TYPE_NAMES = ['TestingPostPutPatchSno', 'TestingDownload']


def get_parameterized_names():
    """ Get all item types from schema names """
    return [name.split('.')[0] for name in os.listdir(os.path.dirname(__file__) + '/../test_schemas')
            if 'mixins' not in name]


PARAMETERIZED_NAMES = get_parameterized_names()


@pytest.mark.parametrize('use_nested', [False, True])
def test_nested_embed(testapp, use_nested):
    with mappings_use_nested(use_nested):
        for uuid in NESTED_OBJECT_LINK_TARGET_GUIDS:
            td_res = testapp.post_json('/nested-object-link-target/',
                                       {
                                           "associates": [{"x": "1", "y": "2", "z": "3"}],
                                           "title": "required title",
                                           "description": "foo",
                                           "uuid": uuid,
                                       },
                                       status=201).json
            assert td_res['@graph'][0]['uuid'] == uuid

        res = testapp.post_json('/nested-embedding-container/',
                                   {
                                       "link_to_nested_object": td_res['@graph'][0]['uuid'],
                                       "title": "required title",
                                       "description": "foo",
                                       "uuid": NESTED_EMBEDDING_CONTAINER_GUID,
                                   },
                                   status=201).json
        embedded_json = testapp.get(res['@graph'][0]['@id'] + "?frame=embedded").json
        print(json.dumps(embedded_json, indent=2))

        link_to_nested_object = embedded_json['link_to_nested_object']
        assert 'associates' in link_to_nested_object
        [associate] = link_to_nested_object['associates']
        assert associate['x'] == '1'
        assert associate['y'] == '2'
        assert 'z' not in associate


# @pytest.mark.xfail
@pytest.mark.parametrize('use_nested', [False, True])
def test_nested_embed_calculated(testapp, use_nested):
    with mappings_use_nested(use_nested):
        for uuid in NESTED_OBJECT_LINK_TARGET_GUIDS:
            td_res = testapp.post_json('/nested-object-link-target/',
                                       {
                                           "associates": [{"x": "1", "y": "2", "z": "3"}],
                                           "title": "required title",
                                           "description": "foo",
                                           "uuid": uuid,
                                       },
                                       status=201).json
            assert td_res['@graph'][0]['uuid'] == uuid

        res = testapp.post_json('/nested-embedding-container/',
                                   {
                                       "link_to_nested_object": NESTED_OBJECT_LINK_TARGET_GUID_1,
                                       "title": "required title",
                                       "description": "foo",
                                       "uuid": NESTED_EMBEDDING_CONTAINER_GUID,
                                   },
                                   status=201).json
        embedded_json = testapp.get(res['@graph'][0]['@id'] + "?frame=embedded").json
        print(json.dumps(embedded_json, indent=2))

        link_to_nested_object = embedded_json['link_to_nested_object']
        assert link_to_nested_object['uuid'] == NESTED_OBJECT_LINK_TARGET_GUID_1
        assert 'associates' in link_to_nested_object
        [associate] = link_to_nested_object['associates']
        assert associate['x'] == '1'
        assert associate['y'] == '2'
        assert 'z' not in associate

        nested_calculated_property = embedded_json['nested_calculated_property']
        assert len(nested_calculated_property) == len(NESTED_OBJECT_LINK_TARGET_GUIDS)
        for expected_uuid, item in zip(NESTED_OBJECT_LINK_TARGET_GUIDS, nested_calculated_property):
            assert item['uuid'] == expected_uuid
            assert 'associates' in item


@pytest.mark.parametrize('use_nested', [False, True])
def test_nested_embed_multi_trivial(testapp, use_nested):
    with mappings_use_nested(use_nested):
        for uuid in NESTED_OBJECT_LINK_TARGET_GUIDS:
            td_res = testapp.post_json('/nested-object-link-target/',
                                       {
                                           "associates": [{"x": "1", "y": "2", "z": "3"}],
                                           "title": "required title",
                                           "description": "foo",
                                           "uuid": uuid,
                                       },
                                       status=201).json
            assert td_res['@graph'][0]['uuid'] == uuid

        res = testapp.post_json('/nested-embedding-container/',
                                   {
                                       "link_to_nested_objects": [td_res['@graph'][0]['uuid']],
                                       "title": "required title",
                                       "description": "foo",
                                       "uuid": NESTED_EMBEDDING_CONTAINER_GUID,
                                   },
                                   status=201).json
        embedded_json = testapp.get(res['@graph'][0]['@id'] + "?frame=embedded").json
        print(json.dumps(embedded_json, indent=2))
        [item] = embedded_json['link_to_nested_objects']  # We only associated one uuid in our post
        [associate] = item['associates'] # We only described one associate in the target
        assert associate['x'] == '1'
        assert associate['y'] == '2'
        assert 'z' not in associate


@pytest.mark.parametrize('use_nested', [False, True])
def test_nested_embed_multi(testapp, use_nested):
    with mappings_use_nested(use_nested):
        minimum_n = len(NESTED_OBJECT_LINK_TARGET_GUIDS)
        n_targets = 5
        assert n_targets >= minimum_n
        props = ["x", "y", "z"]
        embedded_props = ["x", "y"]
        unembedded_props = sorted(set(props) - set(embedded_props))
        assert unembedded_props == ["z"]  # a consistency check. needs to change if props or embedded props changes
        props_count = len(props)  # unless test changes, this is probably 3
        assert props_count == 3  # another consistency check. this needs to change if props changes.

        target_uuid_list = []
        for i in range(n_targets):
            # k will step by the number of properties (probably 3, for "x", "y", "z", but we'll keep it abstract)
            k = i * props_count
            # e.g., prop_bindings will take on {"x": "0", "y": "1", "z": "2"}, {"x": "3", "y": "4", "z": "5"}, ...
            prop_bindings = {key: str(k + pos) for pos, key in enumerate(props)}
            target_result = testapp.post_json('/nested-object-link-target/',
                                              {
                                                  "associates": [prop_bindings],
                                                  # Fields "title" and "description" are required, but might as well
                                                  # include useful stuff in case it's needed for debugging. These values
                                                  # do not affect the outcome. -kmp 4-Jul-2020
                                                  "title": "Item {i} with props={props}".format(i=i, props=prop_bindings),
                                                  "description": "This is item {}".format(i),
                                                  # Here we make sure that we make sure that the various
                                                  "uuid": (NESTED_OBJECT_LINK_TARGET_GUIDS[i]
                                                           if i < minimum_n
                                                           else str(uuid4())),
                                              },
                                              status=201).json
            uuid = target_result['@graph'][0]['uuid']
            target_uuid_list.append(uuid)
        res = testapp.post_json('/nested-embedding-container/',
                                {
                                    "link_to_nested_objects": target_uuid_list,
                                    "title": "Sample container for nested items",
                                    "description": "This is just an example for testing.",
                                    "uuid": NESTED_EMBEDDING_CONTAINER_GUID,
                                },
                                status=201).json
        embedded_json = testapp.get(res['@graph'][0]['@id'] + "?frame=embedded").json
        # This will show up if debugging is needed
        print(json.dumps(embedded_json, indent=2))
        items = embedded_json['link_to_nested_objects']
        assert len(items) == n_targets
        for i, item in enumerate(items):
            [associate] = item['associates']
            # Make sure that "x" and "y" (or whatever our embedded props are) exist in the example.
            assert all(isinstance(associate[prop], str) for prop in embedded_props)
            # Make sure that the fields we did NOT embed do not appear
            assert all(prop not in associate for prop in unembedded_props)
            # Get the values of each of the fields, which we expect to be strings coerceable to integers.
            nums = [int(associate[prop]) for prop in embedded_props]
            # for some item i, the numbers we'll expect are i*prop_count+pos
            k = i * props_count   # Truncate to block of 3. e.g., (0,1,2) or (9,10,11)
            for pos, prop in enumerate(embedded_props):
                assert nums[pos] == k + pos


@pytest.mark.parametrize('item_type', PARAMETERIZED_NAMES)
def test_collections(testapp, item_type):
    """ Get all item types, check they are in the response """
    res = testapp.get('/' + item_type).follow(status=200)
    assert item_type.encode('utf-8') in res.body


@pytest.mark.parametrize('item_type', PARAMETERIZED_NAMES)
def test_json(testapp, item_type):
    """ Check that when we get proper item types """
    res = testapp.get('/' + item_type).follow(status=200)
    assert (item_type + 'Collection') in res.json['@type']


def test_json_basic_auth(anontestapp):
    url = '/'
    value = "Authorization: Basic %s" % ascii_native_(b64encode(b'nobody:pass'))
    res = anontestapp.get(url, headers={'Authorization': value}, status=401)
    assert res.content_type == 'application/json'


def test_home_json(testapp):
    res = testapp.get('/', status=200)
    assert res.json['@type']


def test_collection_post_bad_json(testapp):
    item = {'foo': 'bar'}
    res = testapp.post_json('/embedding-tests', item, status=422)
    assert res.json['status'] == 'error'


def test_collection_post_malformed_json(testapp):
    item = '{'
    headers = {'Content-Type': 'application/json'}
    res = testapp.post('/embedding-tests', item, status=400, headers=headers)
    assert res.json['detail'].startswith('Expecting')


def test_collection_post_missing_content_type(testapp):
    item = '{}'
    testapp.post('/embedding-tests', item, status=415)


def test_collection_post_bad(anontestapp):
    value = "Authorization: Basic %s" % ascii_native_(b64encode(b'nobody:pass'))
    anontestapp.post_json('/embedding-tests', {}, headers={'Authorization': value}, status=401)


def test_collection_limit(testapp):
    """ Post 3 EmbeddingTests, check that limit=all, limit=2 works """
    obj1 = {
        'title': "Testing1",
        'description': "This is testing object 1",
    }
    obj2 = {
        'title': "Testing2",
        'description': "This is testing object 2",
    }
    obj3 = {
        'title': "Testing3",
        'description': "This is testing object 3",
    }
    testapp.post_json('/embedding-tests', obj1, status=201)
    testapp.post_json('/embedding-tests', obj2, status=201)
    testapp.post_json('/embedding-tests', obj3, status=201)
    res_all = testapp.get('/embedding-tests/?limit=all', status=200)
    res_2 = testapp.get('/embedding-tests/?limit=2', status=200)
    assert len(res_all.json['@graph']) == 3
    assert len(res_2.json['@graph']) == 2


def test_collection_put(testapp, execute_counter):
    """ Insert and udpate an item into a collection, verify it worked """
    initial = {
        'title': "Testing",
        'type': "object", # include a non-required field
        'description': "This is the initial insert",
    }
    item_url = testapp.post_json('/embedding-tests', initial).location

    with execute_counter.expect(1):
        item = testapp.get(item_url).json

    for key in initial:
        assert item[key] == initial[key]

    update = {
        'title': "New Testing",
        'type': "object",
        'description': "This is the updated insert",
    }
    testapp.put_json(item_url, update, status=200)

    res = testapp.get('/' + item['uuid']).follow().json

    for key in update:
        assert res[key] == update[key]


def test_invalid_collection_put(testapp):
    """ Tests that inserting various invalid items will appropriately fail """
    missing_required = {
        'title': "Testing",
        'type': "object"
    }
    testapp.post_json('/embedding-tests', missing_required, status=422)

    nonexistent_field = {
        'title': "Testing",
        'type': "string",
        'descriptionn': "This is a descriptionn", # typo
    }
    testapp.post_json('/embedding-tests', nonexistent_field, status=422)

    valid = {
        'title': "Testing",
        'type': "object",
        'description': "This is a valid object",
    }
    invalid_update = {
        'descriptionn': "This is an invalid update",
    }
    item_url = testapp.post_json('/embedding-tests', valid, status=201).location
    testapp.put_json(item_url, invalid_update, status=422)


def test_page_toplevel(anontestapp):
    res = anontestapp.get('/embedding-tests/', status=200)
    assert res.json['@id'] == '/embedding-tests/'


def test_jsonld_context(testapp):
    res = testapp.get('/terms/', status=200)
    assert res.json


@pytest.mark.slow
@pytest.mark.parametrize('item_type', PARAMETERIZED_NAMES)
def test_index_data_workbook(testapp, indexer_testapp, item_type):
    res = testapp.get('/%s?limit=all' % item_type).follow(status=200)
    for item in res.json['@graph']:
        indexer_testapp.get(item['@id'] + '@@index-data', status=200)


def test_home(testapp):
    testapp.get('/', status=200)


@pytest.mark.parametrize('item_type', TYPE_NAMES)
def test_profiles(testapp, item_type):
    res = testapp.get('/profiles/%s.json' % item_type).maybe_follow(status=200)
    errors = Draft202012Validator.check_schema(res.json)
    assert not errors
    # added from ..schema_views._annotated_schema
    assert 'rdfs:seeAlso' in res.json
    assert 'rdfs:subClassOf' in res.json
    assert 'children' in res.json
    assert res.json['isAbstract'] is False


@pytest.mark.parametrize('item_type', ['AbstractItemTest'])
def test_profiles_abstract(testapp, item_type):
    res = testapp.get('/profiles/%s.json' % item_type).maybe_follow(status=200)
    errors = Draft202012Validator.check_schema(res.json)
    assert not errors
    # added from ..schema_views._annotated_schema
    assert 'rdfs:seeAlso' in res.json
    # Item/item does not have subClass
    if item_type.lower() == 'item':
        assert 'rdfs:subClassOf' not in res.json
    else:
        assert 'rdfs:subClassOf' in res.json
    # abstract types wil have children
    assert len(res.json['children']) >= 1
    assert res.json['isAbstract'] is True


def test_profiles_all(testapp, registry):
    res = testapp.get('/profiles/').maybe_follow(status=200)
    # make sure all types are present, including abstract types
    for ti in registry[TYPES].by_item_type.values():
        assert ti.name in res.json
    for ti in registry[TYPES].by_abstract_type.values():
        assert ti.name in res.json


def test_item_revision_history(testapp, registry):
    """ Posts an item then patches it a few times, verifies that all revisions show up in
        the revision history.
    """
    objv1 = {
        'title': "Testing1",
        'description': "This is testing object 1",
    }
    objv2 = {
        'title': "Testing2",
        'description': "This is testing object 2",
    }
    objv3 = {
        'title': "Testing3",
        'description': "This is testing object 3",
    }
    item_uuid = testapp.post_json('/embedding-tests', objv1, status=201).json['@graph'][0]['uuid']
    testapp.patch_json('/' + item_uuid, objv2, status=200)
    testapp.patch_json('/' + item_uuid, objv3, status=200)

    # now get revision history
    revisions = testapp.get('/' + item_uuid + '/@@revision-history').json['revisions']
    assert len(revisions) == 3  # we made 3 edits

    # lets make some more
    testapp.patch_json('/' + item_uuid, objv2, status=200)
    testapp.patch_json('/' + item_uuid, objv1, status=200)
    revisions = testapp.get('/' + item_uuid + '/@@revision-history').json['revisions']
    assert len(revisions) == 5  # now we made 5 edits
    # they should be ordered by sid, recall the patch order above
    for patched_metadata, revision in zip([objv1, objv2, objv3, objv2, objv1], revisions):
        assert revision['title'] == patched_metadata['title']


def _test_auth_config(testapp, registry):
    cfg = testapp.get('/auth0_config').json
    assert cfg['title'] == 'Auth0 Config'
    assert cfg['auth0Client'] == registry.settings['auth0.client']
    assert cfg['auth0Domain'] == registry.settings['auth0.domain']
    assert cfg['auth0Options'] == registry.settings['auth0.options']


def test_auth0_config_admin(testapp, registry):
    """ Tests that acquiring auth0 config gives the expected values from settings for admins. """
    _test_auth_config(testapp, registry)


def test_auth0_config_anon(anontestapp, registry):
    """ Tests that acquiring auth0 config gives the expected values from settings for anonymous users. """
    _test_auth_config(anontestapp, registry)


def _test_recaptcha_config(testapp, registry):
    cfg = testapp.get('/recaptcha_config').json
    assert cfg['title'] == 'Recaptcha Config'
    assert cfg['RecaptchaKey'] == registry.settings['g.recaptcha.key']


def test_recaptcha_config_admin(testapp, registry):
    """ Tests that acquiring recaptcha config gives the expected values from settings for admins. """
    _test_recaptcha_config(testapp, registry)


def test_recaptcha_config_anon(anontestapp, registry):
    """ Tests that acquiring recaptcha config gives the expected values from settings for anonymous users. """
    _test_recaptcha_config(anontestapp, registry)



import json
# import os
import urllib.parse

from pyramid.httpexceptions import HTTPBadRequest
# from pyramid.request import Request
from pyramid.view import view_config
from ..interfaces import TYPES

from ..util import debug_log, get_item_or_none
from ..types.filter_set import FLAGS, FILTER_BLOCKS

from .lucene_builder import LuceneBuilder
from .search import SearchBuilder, search as single_query_search
from .search_utils import execute_search, build_sort_dicts, make_search_subreq


def includeme(config):
    config.add_route('compound_search', '/compound_search')
    config.add_route('build_query', '/build_query{slash:/?}')
    config.scan(__name__)


class CompoundSearchBuilder:
    """ Encapsulates methods needed to run a compound search, in other words an
        AND or an OR query combining a set of queries.

        Entry point is "execute_filter_set".
    """
    TYPE = 'search_type'
    ID = '@id'
    QUERY = 'query'
    NAME = 'name'
    FLAGS_APPLIED = 'flags_applied'
    BUILD_QUERY_URL = '/build_query/'

    @staticmethod
    def build_subreq_from_single_query(request, query, route='/search/', from_=0, to=10):
        """ Builds a Request object that is a proper sub-request of the given request.
            Passes flags directly as query string params. Intended for use with search.

        :param request: request to build off of
        :param query: search query
        :param route: route of sub-request to build
        :param from_: starting ES hit index
        :param to: how many results to returning, index starting at from_
        :return: new Request
        """

        # do some sanitization
        # This will actually urlencode the entire previous string, so:
        # `&inheritance_modes=Neurodev+2500%2B+%282598%29` will become:
        # `&inheritance_modes=Neurodev%2B2500%252B%2B%25282598%2529` (the "&" & "=" chars can be encoded too, but left intact)
        # so the "+" become urlencoded into 'plus-sign' encodings (%2B) (rather than space - %20), and previous 'plus-sign'
        # encodings go from `%2B` to `%252B` (the percent sign gets encoded). So essentially is an encoding of an encoding.
        # But it works once goes into snovault's `make_subreq` downstream.
        if len(query) > 0 and query[0] != "?":
            query = '?' + query

        # If any '?', '&', or '=' in search term, should have been pre-encoded.
        # Meant to handle "+" especially.
        query = urllib.parse.quote(query, safe="?&=")

        subreq = make_search_subreq(request, route + '%s&from=%s&limit=%s' % (query, from_, to))
        subreq.headers['Accept'] = 'application/json'

        return subreq

    @staticmethod
    def combine_query_strings(qstring1, qstring2):
        """ Builds a single URL query from the given flags and blocks.

        :param qstring1: flags, usually ? prefixed
        :param qstring2: blocks to add to it
        :return: combined query
        """

        dict_to_merge_into = dict(urllib.parse.parse_qs(qstring1.lstrip('?'), keep_blank_values=True))
        dict_with_more_vals = dict(urllib.parse.parse_qs(qstring2.lstrip('?'), keep_blank_values=True))

        for k, v in dict_with_more_vals.items():
            if k in dict_to_merge_into:
                dict_to_merge_into[k] += v
            else:
                dict_to_merge_into[k] = v

        return urllib.parse.urlencode(dict_to_merge_into, doseq=True)

    @staticmethod
    def format_result_for_endpoint_response(request, es_results, filter_set, result_sort, search_builder_instance):
        """ Formats es_results from filter_set into a dictionary containing total and @graph,
            setting status on the request if needed. Also sets "__matching_filter_block_names" computed field to identify
            which filter block indices the result matched.

        :param request: current request
        :param es_results: response from ES
        :return: dictionary response
        """

        if es_results['hits']['total']['value'] == 0:
            request.response.status_code = 404  # see google webmaster doc on why

        if search_builder_instance.search_session_id:  # Is 'None' if e.g. limit=all
            request.response.set_cookie('searchSessionID', search_builder_instance.search_session_id)

        result_list = []
        for hit in es_results['hits'].get("hits", []):
            result = hit['_source']['embedded']
            # Matched query names are returned as string here, regardless of their specified original type.
            result["__matching_filter_block_names"] = hit.get("matched_queries", [])
            result_list.append(result)

        columns = SearchBuilder.build_initial_columns([ request.registry[TYPES][filter_set[CompoundSearchBuilder.TYPE]].schema ])
        # We used multiple filter blocks, so we add in column for "__matching_filter_block_names"
        columns["__matching_filter_block_names"] = {
            "title": "Filter Blocks Matched",
            "order": 1000
        }

        return {
            # "@id": "/compound_search", # Removed - presense of @id on UI is inferred to mean that there is 1 filter block in request.
            # "@type": ["SearchResults"], # Not necessary from UI atm but can consider adding for semantics
            "total": es_results['hits'].get("total", {}).get("value", 0),
            "@graph": result_list,
            "columns": columns,
            "sort": result_sort
        }

    @staticmethod
    def invoke_search(context, request, subreq, search_type, return_generator=False):
        """
        Wrapper method that invokes the core search API (/search/) with the given subreq and
        copies over status code to parent response.

        :param context: context of parent request
        :param request: parent request
        :param subreq: subrequest
        :return: response from /search/
        """
        # Calls SearchBuilder.format_results internally, incl. adding searchSessionID cookie to response.
        response = single_query_search(context, subreq, search_type, return_generator)
        if subreq.response.status_code == 404:
            request.response.status_code = 404
        return response

    @staticmethod
    def _add_type_to_flag_if_needed(flags, type_flag):
        """ Modifies 'flags' in place by adding type query if it is not present

        :param flags: query substring
        :param type_flag: query substring containing type requirement
        :return: query string that combines the two, if type requirement isn't already there
        """
        if type_flag not in flags or type_flag.lower() not in flags:
            if len(flags) > 0:
                flags += '&' + type_flag
            else:
                flags = type_flag

        return flags

    @staticmethod
    def es_results_generator(es_results):
        for hit in es_results['hits'].get('hits', []):
            yield hit['_source']['embedded']

    @staticmethod
    def execute_filter_set(context, request, filter_set, from_=0, to=10,
                           global_flags=None, return_generator=False, intersect=False):
        """ Executes the given filter_set. This function contains the core functionality of the class.
            A filter_set with respect to this function is just a dictionary containing the following things:
                1. 'search_type' is the item type we are executing on. Required.
                2. 'filter_blocks' contains the filter blocks we would like to apply with named flags we
                    would like to apply on this block as well
                3. 'flags' is a dictionary containing named flags to be applied to individual filter_blocks
                    by name.

                NOTE: if neither 'flags' nor 'filter_blocks' is specified then a generic type=Item
                search will be executed. If just 'flags' is specified with no filter_blocks, the
                flags will be ignored (since there are no filter_blocks to apply it to).
        """
        cls = CompoundSearchBuilder
        filter_blocks = filter_set.get(FILTER_BLOCKS, [])
        flags = filter_set.get(FLAGS, None)
        doc_type = filter_set.get(CompoundSearchBuilder.TYPE)
        search_type = filter_set.get(cls.TYPE, 'Item')  # if type not set, attempt to search on item
        type_flag = 'type=%s' % search_type

        # if we have no filter blocks, there is no context to enable flags, so
        # pass type_flag + global_flags
        if not filter_blocks and flags:
            if global_flags:
                query = cls.combine_query_strings(global_flags, type_flag)
            else:
                query = type_flag
            subreq = cls.build_subreq_from_single_query(request, query, from_=from_, to=to)
            return CompoundSearchBuilder.invoke_search(context, request, subreq, search_type, return_generator)

        # if we specified global_flags, combine that query with the single filter_block,
        # otherwise pass the filter_block query directly
        elif not flags and len(filter_blocks) == 1:
            block = filter_blocks[0]
            block_query = block[cls.QUERY]
            if global_flags:
                query = cls.combine_query_strings(global_flags, block_query)
            else:
                query = block_query
            query = cls._add_type_to_flag_if_needed(query, type_flag)
            subreq = cls.build_subreq_from_single_query(request, query, from_=from_, to=to)
            return CompoundSearchBuilder.invoke_search(context, request, subreq, search_type, return_generator)

        # Extract query string and list of applied flags, add global_flags to block_query first
        # then add flags as applied and type_flag if needed.
        elif flags and len(filter_blocks) == 1:
            block_query = filter_blocks[0][cls.QUERY]
            flags_applied = filter_blocks[0][cls.FLAGS_APPLIED]
            if global_flags:
                query = cls.combine_query_strings(global_flags, block_query)
            else:
                query = block_query
            for applied_flag in flags_applied:
                for flag in flags:
                    if flag[cls.NAME] == applied_flag:
                        query = cls.combine_query_strings(query, flag[cls.QUERY])
                        break
            query = cls._add_type_to_flag_if_needed(query, type_flag)
            subreq = cls.build_subreq_from_single_query(request, query, from_=from_, to=to)
            return CompoundSearchBuilder.invoke_search(context, request, subreq, search_type, return_generator)

        # Build the compound_query
        # Iterate through filter_blocks, adding global_flags if specified and adding flags if specified
        else:
            sub_queries = []
            for block_index, block in enumerate(filter_blocks):
                block_query = block[cls.QUERY]
                flags_applied = block[cls.FLAGS_APPLIED]
                query = block_query
                if global_flags:
                    query = cls.combine_query_strings(global_flags, block_query)
                for applied_flag in flags_applied:
                    for flag in flags:
                        if flag[cls.NAME] == applied_flag:
                            query = cls.combine_query_strings(query, flag[cls.QUERY])
                            break
                query = cls._add_type_to_flag_if_needed(query, type_flag)
                subreq = cls.build_subreq_from_single_query(request, query, route=cls.BUILD_QUERY_URL,
                                                            from_=from_, to=to)
                sub_query = request.invoke_subrequest(subreq).json[cls.QUERY]
                # See https://www.elastic.co/guide/en/elasticsearch/reference/7.17/query-dsl-bool-query.html#named-queries
                sub_query["bool"]["_name"] = str(block.get("name", block_index))  # note in ES7 numbers here must be cast to string
                sub_queries.append(sub_query)


            compound_query = LuceneBuilder.compound_search(sub_queries, intersect=intersect)
            compound_subreq = cls.build_subreq_from_single_query(request, ('?type=' + search_type))

            requested_sorts = filter_set.get("sort", [])
            if not requested_sorts and global_flags:
                requested_sorts = urllib.parse.parse_qs(global_flags).get("sort", [])

            sort, result_sort = build_sort_dicts(requested_sorts, request, [ doc_type ])

            search_builder_instance = SearchBuilder.from_search(context, compound_subreq, compound_query, return_generator=return_generator)
            search_builder_instance.assure_session_id()
            search_builder_instance.query['sort'] = sort
            es_results = None
            if return_generator and to == "all":
                es_results = search_builder_instance.execute_search_for_all_results()
                # es_results['hits']['hits'] WILL ALREADY BE A GENERATOR/ITERATOR -- SEE search.py:execute_search_for_all_results
                # (Unless result count is <100, in which case will be a list)
                return cls.es_results_generator(es_results)
            elif return_generator and to != "all":
                # TODO: Consider enabling somehow
                raise HTTPBadRequest("`return_generator` without `limit=all` not yet supported")
            elif to == "all":
                raise HTTPBadRequest("`limit=all` is not permitted without `return_generator`")
            else:
                es_results = execute_search(
                    es=search_builder_instance.es,
                    query=search_builder_instance.query,
                    index=search_builder_instance.es_index,
                    from_=from_,
                    size=to,
                    session_id=search_builder_instance.search_session_id
                )
                return cls.format_result_for_endpoint_response(request, es_results, filter_set, result_sort, search_builder_instance)

    @staticmethod
    def validate_flag(flag):
        """ Validates a given flag has the correct structure/types """
        if CompoundSearchBuilder.NAME not in flag or CompoundSearchBuilder.QUERY not in flag:  # existence
            raise HTTPBadRequest('Passed a bad flag with missing structure: %s' % flag)
        elif not isinstance(flag[CompoundSearchBuilder.NAME], str):  # type
            raise HTTPBadRequest('Passed a bad flag with incorrect parameter for field %s: %s' % (CompoundSearchBuilder.NAME, flag))
        elif not isinstance(flag[CompoundSearchBuilder.QUERY], str):  # type
            raise HTTPBadRequest('Passed a bad flag with incorrect parameter for field %s: %s' % (CompoundSearchBuilder.QUERY, flag))

    @staticmethod
    def validate_filter_block(filter_block):
        """ Validates a given filter_block has correct structure/types """
        if CompoundSearchBuilder.QUERY not in filter_block or CompoundSearchBuilder.FLAGS_APPLIED not in filter_block:
            raise HTTPBadRequest('Passed a bad filter_block with missing structure: %s' % filter_block)
        elif not isinstance(filter_block[CompoundSearchBuilder.QUERY], str):
            raise HTTPBadRequest('Passed a bad filter_block with wrong type for field %s: %s' % (CompoundSearchBuilder.QUERY, filter_block))
        elif not isinstance(filter_block[CompoundSearchBuilder.FLAGS_APPLIED], list):
            raise HTTPBadRequest('Passed a bad filter_block with wrong type for field %s: %s' %
                                 (CompoundSearchBuilder.FLAGS_APPLIED, filter_block))

    @staticmethod
    def extract_filter_set_from_search_body(request, body):
        """ Validates the compound_search POST request body, returning a dictionary filter_set item.

        :param request: current request
        :param body: body of POST request (in JSON)
        :return: a filter_set, to be executed
        """
        if CompoundSearchBuilder.ID in body:  # prioritize @id
            return get_item_or_none(request, body[CompoundSearchBuilder.ID])
        else:
            filter_set = {}
            if CompoundSearchBuilder.TYPE in body:
                filter_set[CompoundSearchBuilder.TYPE] = body[CompoundSearchBuilder.TYPE]
            else:
                raise HTTPBadRequest('Tried to execute a filter_set without specifying a type!')
            if FLAGS in body:
                if not isinstance(body[FLAGS], list):
                    raise HTTPBadRequest('Passed a bad value for flags: %s -- Expected a list.' % body[FLAGS])
                for flag in body[FLAGS]:
                    CompoundSearchBuilder.validate_flag(flag)
                filter_set[FLAGS] = body[FLAGS]
            if FILTER_BLOCKS in body:
                if not isinstance(body[FILTER_BLOCKS], list):
                    raise HTTPBadRequest('Passed a bad value for flags: %s -- Expected a list.' % body[FILTER_BLOCKS])
                for filter_block in body[FILTER_BLOCKS]:
                    CompoundSearchBuilder.validate_filter_block(filter_block)
                filter_set[FILTER_BLOCKS] = body[FILTER_BLOCKS]
            return filter_set





@view_config(route_name='build_query', request_method='GET', permission='search')
@debug_log
def build_query(context, request):
    """ Runs the query construction step of the search, returning the lucene query as the response.
        Used as a helper for compound_search, making 1 sub-request per filter_block.
    """
    builder = SearchBuilder(context, request)
    builder._build_query()
    return builder.query


@view_config(route_name='compound_search', request_method='POST', permission='search')
@debug_log
def compound_search(context, request):
    """ Executes a compound_search given a uuid of a filter_set (or filter_set props, tbd).

        You have two options when executing a compound search - you can pass a uuid of an existing
        filter_set item or you can pass the relevant filter_set fields directly. This allows the
        client to acquire/cache filter_sets then pass modified query params directly to ES without
        triggering a write to the base filter_set.

        POST Body Syntax:
        {
            # flags to be applied globally to the search
            "global_flags": <query_string>,

            # uuid of a filter_set item to execute
            "uuid": <uuid>,  # NOTE: if you provide this, the following filter_set related fields are IGNORED

            "search_type": <item_type>,  # item type this filter_set is searching on
            "flags": [
                {
                    "name": "flag_name_one",
                    "query": <query_string>
                },
                {
                    "name": "flag_name_two",
                    "query": <query_string>
                }
                ...
            ]

            # list of queries to be compounded with below structure
            "filter_blocks": [
                {
                    "query": <query_string>, (to be combined with global_flags, if specified)
                    "flags_applied": [
                        "flag_name_one",
                        "flag_name_two"
                    ]
                }
                ...
            ]

            # other options
            "from": <int>,  # starting index in ES search results to return, default 0
            "limit": <int>,  # number of results to return, default 25
            "return_generator": true/false, default false
            "intersect": true/false, if true will compute intersection of filter_blocks, default false
        }

    """
    body = json.loads(request.body)

    filter_set = CompoundSearchBuilder.extract_filter_set_from_search_body(request, body)
    global_flags = body.get('global_flags', None)
    intersect = True if body.get('intersect', False) else False

    # Disabled for time being to allow test(s) to pass. Not sure whether to add Project to FilterSet schema 'search_type' enum.
    # if filter_set.get(CompoundSearchBuilder.TYPE) not in request.registry[TYPES]["FilterSet"].schema["properties"][CompoundSearchBuilder.TYPE]["enum"]:
    #    raise HTTPBadRequest("Passed bad {} body param: {}".format(CompoundSearchBuilder.TYPE, filter_set.get(CompoundSearchBuilder.TYPE)))

    from_ = body.get('from', 0)
    limit = body.get('limit', 10)  # pagination size 10 works better with ECS
    if limit == "all":
        raise HTTPBadRequest("compound_search does not support limit=all at this time.")
    if limit > 1000:
        limit = 1000
    if from_ < 0 or limit < 0:
        raise HTTPBadRequest('Passed bad from, to request body params: %s, %s' % (from_, limit))

    return CompoundSearchBuilder.execute_filter_set(
        context,
        request,
        filter_set,
        from_=from_,
        to=limit,
        global_flags=global_flags,
        intersect=intersect,
        return_generator=False
    )



from __future__ import annotations

import argparse
import itertools
import json
import logging
import re
import structlog
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Iterable, Iterator, List, Optional, Set, Union

from dcicutils import creds_utils
from dcicutils.ff_utils import get_metadata, search_metadata
from dcicutils.misc_utils import to_snake_case

from snovault.commands.utils import get_auth_key


logger = structlog.getLogger(__name__)
EPILOG = __doc__

DEFAULT_IGNORE_FIELDS = [
    "submitted_by",
    "date_created",
    "last_modified",
    "schema_version",
]
MASTER_INSERTS = "master-inserts"
INSERT_DIRECTORIES = [
    "inserts",
    MASTER_INSERTS,
    "perf-testing",
    "workbook-inserts",
    "temp-local-inserts",
    "deploy-inserts",
]
INSERTS_LOCATION = Path("src/encoded/tests/data/")
KEYS_TO_IGNORE_FOR_LINKS = set(  # Either UUIDs not needed or can look like UUIDs
    ["uuid", "title", "blob_id", "md5sum", "content_md5sum", "submitted_md5sum"]
)
UUID_IDENTIFIER = re.compile(
    r"^[a-f0-9]{8}-?[a-f0-9]{4}-?4[a-f0-9]{3}-?[a-f0-9]{4}-?[a-f0-9]{12}\Z",
    re.I,
)


def get_ignore_fields(ignore_fields: List[str]) -> Set[str]:
    """Get all fields to ignore when pulling inserts."""
    return set(DEFAULT_IGNORE_FIELDS + ignore_fields)


def update_inserts_from_server(
    inserts: Path,
    auth_key: Dict[str, str],
    ignore_fields: Iterable[str],
    item_types: Optional[List[str]] = None,
    update_existing: bool = False,
    from_search: Optional[str] = None,
    merge_existing: bool = True,
) -> None:
    """Update inserts for given server."""
    existing_inserts = get_existing_inserts(inserts)
    logger.info(f"Found existing inserts for {len(existing_inserts)} item types")
    existing_uuids = get_existing_uuids_to_update(
        existing_inserts, item_types=item_types, update_all=update_existing
    )
    if existing_uuids:
        logger.info(f"Found {len(existing_uuids)} existing items to update")
    search_uuids = get_uuids_from_search(from_search, auth_key)
    if search_uuids:
        logger.info(f"Found {len(search_uuids)} items from search")
    base_uuids = get_base_uuids(existing_uuids, search_uuids)
    logger.info("Collecting inserts from portal. This may take a while...")
    inserts_from_portal = get_inserts_from_portal(base_uuids, auth_key, ignore_fields)
    logger.info(
        f"Found inserts for {len(inserts_from_portal)} item types from portal:"
        f" {[item_type_inserts.item_type for item_type_inserts in inserts_from_portal]}"
    )
    inserts_to_write = get_inserts_to_write(
        inserts, inserts_from_portal, existing_inserts, merge_existing=merge_existing
    )
    logger.info(f"Writing inserts for {len(inserts_to_write)} item types to {inserts}")
    write_inserts(inserts_to_write, inserts)


@dataclass(frozen=True)
class Insert:
    uuid: str
    properties: Dict[str, Any]

    def __hash__(self):
        return hash(self.uuid)

    def __eq__(self, other: Any):
        if not isinstance(other, Insert):
            return False
        return self.uuid == other.uuid

    def __repr__(self):
        return f"Insert({self.uuid})"

    def update(
        self,
        properties: Optional[Dict[str, Any]] = None,
    ) -> Insert:
        """Update insert attributes."""
        if not properties or properties == self.properties:
            return self
        return Insert(
            uuid=self.uuid,
            properties=properties,
        )


@dataclass(frozen=True)
class ItemTypeInserts:
    item_type: str
    uuids: Set[str]
    inserts: Iterator[Insert]

    def __repr__(self):
        return f"{self.__class__.__name__}({self.item_type})"

    def add_insert(self, insert: Insert) -> ItemTypeInserts:
        """Update inserts for a given item type."""
        if insert.uuid in self.uuids:
            return self
        return ItemTypeInserts(
            item_type=self.item_type,
            uuids=self.uuids | {insert.uuid},
            inserts=itertools.chain(self.inserts, [insert]),
        )


def get_existing_uuids_to_update(
    existing_inserts: List[ItemTypeInserts],
    item_types: Optional[List[str]] = None,
    update_all: bool = False,
) -> Set[str]:
    """Get all existing inserts to update from given directory.

    Note: Pulling existing inserts into memory
    """
    if update_all:
        return set(
            itertools.chain(
                *[item_type_inserts.uuids for item_type_inserts in existing_inserts]
            )
        )
    return get_uuids_for_item_types(existing_inserts, item_types)


def get_existing_inserts(inserts_path: Path) -> List[ItemTypeInserts]:
    """Get all existing inserts from given directory.

    Ignore any item types with no inserts.
    """
    inserts_files = inserts_path.glob("*.json")
    item_type_inserts = [
        get_item_type_inserts(insert_file) for insert_file in inserts_files
    ]
    return get_non_empty_item_type_inserts(item_type_inserts)


def get_non_empty_item_type_inserts(
    item_type_inserts: List[ItemTypeInserts],
) -> List[ItemTypeInserts]:
    """Get all non-empty item type inserts."""
    return [
        item_type_insert
        for item_type_insert in item_type_inserts
        if item_type_insert.uuids
    ]


def get_item_type_inserts(insert_file: Path) -> ItemTypeInserts:
    """Get all inserts for a given item type.

    Ensure inserts are sorted by UUID for later comparison with portal
    inserts.
    """
    item_type = insert_file.stem
    uuids = set()
    inserts = []
    file_inserts = get_inserts_from_file(insert_file)
    for insert in sort_inserts_by_uuid(file_inserts):
        uuids |= {insert.uuid}
        inserts = itertools.chain(inserts, [insert])
    return ItemTypeInserts(
        item_type=item_type,
        uuids=uuids,
        inserts=inserts,
    )


def create_insert(insert: Dict[str, Any]) -> Insert:
    """Create Insert from JSON."""
    return Insert(
        uuid=get_uuid(insert),
        properties=insert,
    )


def get_inserts_from_file(insert_file: Path) -> Iterator[Insert]:
    """Load inserts from file."""
    with insert_file.open("r") as f:
        inserts = json.load(f)
    return (create_insert(insert) for insert in inserts)


def sort_inserts_by_uuid(inserts: Iterable[Insert]) -> List[Insert]:
    """Sort all inserts by UUID."""
    return sorted(inserts, key=lambda insert: insert.uuid)


def get_uuid(item: Dict[str, Any]) -> str:
    return item.get("uuid", "")


def get_uuids_for_item_types(
    existing_inserts: List[ItemTypeInserts],
    item_types: Optional[List[str]] = None,
) -> Set[str]:
    """Get all inserts to update from given directory."""
    if item_types:
        item_types_to_keep = set([to_snake_case(target) for target in item_types])
        return set(
            itertools.chain(
                *[
                    item_type_inserts.uuids
                    for item_type_inserts in existing_inserts
                    if item_type_inserts.item_type in item_types_to_keep
                ]
            )
        )
    return set()


def get_uuids_from_search(
    search_query: Union[str, None], auth_key: Dict[str, str]
) -> List[str]:
    """Get all uuids from a given search query."""
    if not search_query:
        return []
    query = format_search_query(search_query)
    return [get_uuid(item) for item in search_metadata(query, key=auth_key)]


def format_search_query(search_query: str) -> str:
    """Format provided search query.

    Include UUID add-on to limit response size and use heuristics to
    format query as needed (e.g. so no need to add boilerplate start of
    "search/").
    """
    if "field=uuid" not in search_query:
        search_query += "&field=uuid"
    if "/" in search_query:  # Likely formatted appropriately
        return search_query
    if not search_query.startswith("?"):
        return f"search/?{search_query}"
    return f"search/{search_query}"


def get_base_uuids(
    existing_uuids: Set[str],
    search_uuids: List[str],
) -> Set[str]:
    """Get all base UUIDs to start pulling inserts from portal."""
    return existing_uuids | set(search_uuids)


def get_inserts_from_portal(
    uuids: Iterable[str], auth_key: Dict[str, str], ignore_fields: Set[str]
) -> List[ItemTypeInserts]:
    """Get all inserts to write.

    Start from base UUIDs and branch to all linked items.
    """
    item_type_inserts = []
    seen = set()
    uuids_to_get = set(uuids)
    while uuids_to_get:
        current_uuids_to_get = uuids_to_get - seen
        uuids_to_get = set()
        for uuid in current_uuids_to_get:
            item_type = get_item_type(uuid, auth_key)
            insert = get_insert(uuid, auth_key, ignore_fields)
            item_type_inserts = update_item_type_inserts(
                item_type_inserts, item_type, insert
            )
            seen |= {uuid}
            uuids_to_get |= get_links(insert) - seen
    return item_type_inserts


def get_item(uuid: str, auth_key: Dict[str, str], frame: str = "raw") -> Dict[str, Any]:
    """Get item for a given UUID."""
    add_on = f"frame={frame}" if frame else ""
    return get_metadata(uuid, key=auth_key, add_on=add_on)


def get_insert(uuid: str, auth_key: Dict[str, str], ignore_fields: Set[str]) -> Insert:
    """Get insert for a given item."""
    item = get_item(uuid, auth_key)
    properties = get_insert_properties(item, ignore_fields)
    return create_insert(properties)


def get_item_type(uuid: str, auth_key: Dict[str, str]) -> str:
    """Get item type for a given item.

    Ensure snake_cased to match insert file names.
    """
    item = get_item(uuid, auth_key, frame="object")
    return to_snake_case(item["@type"][0])


def get_insert_properties(
    item: Dict[str, Any], ignore_fields: Set[str]
) -> Dict[str, Any]:
    """Get all properties for a given item."""
    return {key: value for key, value in item.items() if key not in ignore_fields}


def update_item_type_inserts(
    item_type_inserts: List[ItemTypeInserts],
    item_type: str,
    insert: Insert,
) -> List[ItemTypeInserts]:
    """Update item type inserts with new insert.

    If item type already exists, add insert to existing item type inserts.
    Otherwise, create new item type inserts.
    """
    existing_item_types = {
        item_type_inserts.item_type for item_type_inserts in item_type_inserts
    }
    if item_type in existing_item_types:
        return [
            item_type_insert.add_insert(insert)
            if item_type_insert.item_type == item_type
            else item_type_insert
            for item_type_insert in item_type_inserts
        ]
    new_item_type_inserts = ItemTypeInserts(
        item_type=item_type,
        uuids={insert.uuid},
        inserts=(insert,),
    )
    return item_type_inserts + [new_item_type_inserts]


def get_links(item: Any) -> Set[str]:
    """Get all links for a given item."""
    if isinstance(item, Insert):
        return get_links_from_insert(item)
    if isinstance(item, dict):
        return get_links_from_dict(item)
    if isinstance(item, list):
        return get_links_from_list(item)
    if isinstance(item, str):
        if is_uuid(item):
            return {item}
    return set()


def get_links_from_insert(item: Insert) -> Set[str]:
    """Get all links for a given insert."""
    return get_links_from_dict(item.properties)


def get_links_from_dict(item: Dict[str, Any]) -> Set[str]:
    """Get all links for a given dictionary."""
    return {
        link
        for key, value in item.items()
        if key not in KEYS_TO_IGNORE_FOR_LINKS
        for link in get_links(value)
    }


def get_links_from_list(item: List[Any]) -> Set[str]:
    """Get all links for a given list."""
    return {link for value in item for link in get_links(value)}


def is_uuid(value: str) -> bool:
    """Check if a given value is a UUID."""
    return bool(UUID_IDENTIFIER.match(value))


def get_inserts_to_write(
    inserts_path: Path,
    inserts_from_portal: List[ItemTypeInserts],
    existing_inserts: List[ItemTypeInserts],
    merge_existing: bool = True,
) -> List[ItemTypeInserts]:
    """Get all inserts to write.

    Update portal inserts with existing inserts information, if present,
    and remove conflicts with master-inserts (unless updating them).
    """
    inserts_to_write = get_item_type_inserts_to_write(
        inserts_from_portal, existing_inserts, merge_existing=merge_existing
    )
    if is_master_inserts(inserts_path):
        return inserts_to_write
    return get_inserts_without_conflicts(inserts_to_write)


def is_master_inserts(inserts_path: Path) -> bool:
    """Check if inserts are master-inserts."""
    return inserts_path.stem == MASTER_INSERTS


def get_item_type_inserts_to_write(
    inserts_from_portal: List[ItemTypeInserts],
    existing_inserts_to_update: List[ItemTypeInserts],
    merge_existing: bool = True,
) -> List[ItemTypeInserts]:
    """Update portal inserts with existing inserts data, if present.

    Merge existing properties as directed.
    """
    if not existing_inserts_to_update:
        return inserts_from_portal
    existing_item_types_to_inserts = {
        item_type_inserts.item_type: item_type_inserts
        for item_type_inserts in existing_inserts_to_update
    }
    return [
        get_updated_item_type_inserts(
            portal_item_type_inserts,
            existing_item_types_to_inserts[portal_item_type_inserts.item_type],
            merge_existing=merge_existing,
        )
        if portal_item_type_inserts.item_type in existing_item_types_to_inserts
        else portal_item_type_inserts
        for portal_item_type_inserts in inserts_from_portal
    ]


def get_updated_item_type_inserts(
    inserts_from_portal: ItemTypeInserts,
    existing_inserts: ItemTypeInserts,
    merge_existing: bool = True,
) -> ItemTypeInserts:
    """Combine portal inserts with existing inserts.

    Walk through portal and existing inserts, collecting all inserts to
    write for a given item type.

    Keep all inserts found only from portal or only in existing inserts
    exactly as they are.

    For inserts found in both portal and existing inserts, keep
    non-overlapping existing properties if directed. Otherwise, only
    keep portal insert properties.

    Note: Existing inserts are assumed to be sorted by UUID (performed
    when loaded from file).
    """
    sorted_portal_inserts = (
        insert for insert in sort_inserts_by_uuid(inserts_from_portal.inserts)
    )
    sorted_existing_inserts = existing_inserts.inserts
    uuids = set()
    inserts = []
    portal_insert = get_next_insert(sorted_portal_inserts)
    existing_insert = get_next_insert(sorted_existing_inserts)
    while portal_insert or existing_insert:
        if not portal_insert:
            insert_to_add = existing_insert
            existing_insert = get_next_insert(sorted_existing_inserts)
        elif not existing_insert:
            insert_to_add = portal_insert
            portal_insert = get_next_insert(sorted_portal_inserts)
        elif portal_insert.uuid < existing_insert.uuid:
            insert_to_add = portal_insert
            portal_insert = get_next_insert(sorted_portal_inserts)
        elif portal_insert.uuid > existing_insert.uuid:
            insert_to_add = existing_insert
            existing_insert = get_next_insert(sorted_existing_inserts)
        else:  # portal_insert.uuid == existing_insert.uuid
            insert_to_add = get_updated_insert(
                portal_insert, existing_insert, merge_existing=merge_existing
            )
            portal_insert = get_next_insert(sorted_portal_inserts)
            existing_insert = get_next_insert(sorted_existing_inserts)
        uuids |= {insert_to_add.uuid}
        inserts = itertools.chain(inserts, [insert_to_add])
    return ItemTypeInserts(
        item_type=inserts_from_portal.item_type,
        uuids=uuids,
        inserts=inserts,
    )


def get_next_insert(inserts: Iterator[Insert]) -> Union[Insert, None]:
    """Get next insert from an iterable."""
    return next(inserts, None)


def get_updated_insert(
    portal_insert: Insert,
    existing_insert: Insert,
    merge_existing: bool = True,
) -> Insert:
    """Update existing insert with portal insert properties.

    If directed, keep existing properties not present in portal insert.
    """
    if merge_existing:
        properties = {
            **existing_insert.properties,
            **portal_insert.properties,
        }
        return portal_insert.update(properties=properties)
    return portal_insert


def get_inserts_without_conflicts(
    item_type_inserts: List[ItemTypeInserts],
) -> List[ItemTypeInserts]:
    """Remove all conflicts with master-inserts."""
    master_inserts = get_existing_inserts(INSERTS_LOCATION.joinpath(MASTER_INSERTS))
    master_inserts_item_types = {
        item_type_inserts.item_type: item_type_inserts
        for item_type_inserts in master_inserts
    }
    item_type_inserts_without_conflicts = [
        get_inserts_without_conflicts_for_item_type(
            item_type_inserts, master_inserts_item_types[item_type_inserts.item_type]
        )
        if are_item_type_inserts_present_and_overlapping(
            item_type_inserts, master_inserts_item_types
        )
        else item_type_inserts
        for item_type_inserts in item_type_inserts
    ]
    return get_non_empty_item_type_inserts(item_type_inserts_without_conflicts)


def get_inserts_without_conflicts_for_item_type(
    portal_item_type_inserts: ItemTypeInserts,
    master_item_type_inserts: ItemTypeInserts,
) -> ItemTypeInserts:
    """Remove conflicts with master-inserts for a given item type.

    Assumes both portal and master inserts are sorted by UUID.
    """
    uuids = set()
    inserts = []
    for insert in sort_inserts_by_uuid(portal_item_type_inserts.inserts):
        if is_insert_in_master_inserts(insert, master_item_type_inserts):
            continue
        uuids |= {insert.uuid}
        inserts = itertools.chain(inserts, [insert])
    return ItemTypeInserts(
        item_type=portal_item_type_inserts.item_type,
        uuids=uuids,
        inserts=inserts,
    )


def is_insert_in_master_inserts(
    insert: Insert,
    master_item_type_inserts: ItemTypeInserts,
) -> bool:
    """Check if a given insert is in master-inserts.

    Assumes master inserts are sorted by UUID and portal inserts arrive
    in order as well.
    """
    if insert.uuid in master_item_type_inserts.uuids:
        for master_insert in master_item_type_inserts.inserts:
            if master_insert.uuid == insert.uuid:
                if master_insert.properties == insert.properties:
                    logging.info(
                        f"Skipping {insert.uuid} for"
                        f" {master_item_type_inserts.item_type} as"
                        f" already in master-inserts"
                    )
                else:
                    logging.warning(
                        f" Skipping {insert.uuid} for"
                        f" {master_item_type_inserts.item_type} as"
                        f" conflicts with master-inserts"
                    )
                return True
    return False


def are_item_type_inserts_present_and_overlapping(
    item_type_inserts: ItemTypeInserts,
    comparison_inserts: Dict[str, ItemTypeInserts],
) -> bool:
    """Check if item type inserts are present in and overlap inserts."""
    return item_type_inserts.item_type in comparison_inserts and do_inserts_overlap(
        item_type_inserts, comparison_inserts[item_type_inserts.item_type]
    )


def do_inserts_overlap(
    item_type_inserts_1: ItemTypeInserts,
    item_type_inserts_2: ItemTypeInserts,
) -> bool:
    """Check if portal inserts overlap with existing inserts."""
    return bool(item_type_inserts_1.uuids & item_type_inserts_2.uuids)


def write_inserts(item_type_inserts: List[ItemTypeInserts], inserts_path: Path) -> None:
    """Write all inserts to given directory."""
    for item_type_insert in item_type_inserts:
        write_inserts_for_type(
            item_type_insert.item_type, item_type_insert.inserts, inserts_path
        )


def write_inserts_for_type(
    item_type: str,
    inserts_for_type: Iterable[Insert],
    inserts_path: Path,
) -> None:
    """Write all inserts for a given item type to given directory."""
    insert_file = inserts_path.joinpath(f"{item_type}.json")
    inserts = sort_inserts_by_uuid(inserts_for_type)
    to_write = [insert.properties for insert in inserts]
    with insert_file.open("w") as file_handle:
        json.dump(to_write, file_handle, indent=4)
    logger.info(f"Wrote {len(inserts)} inserts for {item_type} to {insert_file}")


def main():
    """Update inserts from a given portal."""
    logging.basicConfig()
    logging.getLogger("encoded").setLevel(logging.DEBUG)

    parser = argparse.ArgumentParser(
        description="Update Inserts",
        epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "--env", default="data", help="Environment to update from. Defaults to data"
    )
    parser.add_argument(
        "--portal",
        help="Portal to update inserts from. Defaults to smaht",
        choices=creds_utils._KEY_MANAGERS.keys(),
        default="smaht",
    )
    parser.add_argument(
        "--dest",
        default="temp-local-inserts",
        help="Destination inserts directory. Defaults to temp-local-inserts.",
    )
    parser.add_argument(
        "--item",
        nargs="+",
        help="Existing item type(s) (e.g. file_fastq) to update inserts for.",
    )
    parser.add_argument(
        "--update",
        help=(
            "Update all existing inserts with data from portal."
            " Overrides --item flag."
        ),
        action="store_true",
    )
    parser.add_argument(
        "--ignore",
        nargs="+",
        default=DEFAULT_IGNORE_FIELDS,
        help="Properties to ignore when pulling inserts",
    )
    parser.add_argument(
        "--search",
        help=(
            "Query to find new items to add to inserts, e.g."
            " 'type=FileFastq&status=uploaded'."
        ),
        type=str,
    )
    parser.add_argument(
        "--refresh",
        help=(
            "Replace existing inserts with new inserts from portal, i.e. don't keep"
            " properties from existing inserts not present in portal."
        ),
        action="store_true",
    )
    args = parser.parse_args()

    ignore_fields = get_ignore_fields(args.ignore)
    auth_key = get_auth_key(args.portal, args.env)
    if args.dest not in INSERT_DIRECTORIES:
        proceed = input(
            f"Destination {args.dest} not in common choices: {INSERT_DIRECTORIES}."
            f"\nContinue? (y/n): "
        )
        if not proceed.lower().startswith("y"):
            logger.info("Exiting.")
            return
    inserts_path = INSERTS_LOCATION.joinpath(args.dest)
    if not inserts_path.exists():
        inserts_path.mkdir()
    update_inserts_from_server(
        inserts_path,
        auth_key,
        ignore_fields,
        item_types=args.item,
        update_existing=args.update,
        from_search=args.search,
        merge_existing=not args.refresh,
    )


if __name__ == "__main__":
    main()



import argparse
import structlog
import logging

from pyramid.paster import get_app
from snovault.elasticsearch.create_mapping import run as run_create_mapping
from snovault.elasticsearch.create_mapping import reindex_by_type_staggered
from dcicutils.log_utils import set_logging
from dcicutils.env_utils import is_stg_or_prd_env, is_test_env

# override this order in the downstream portal
from ..loadxl import loadxl_order

log = structlog.getLogger(__name__)
EPILOG = __doc__


def get_my_env(app):
    """
    Gets the env name of the currently running environment

    :param app: handle to Pyramid app
    :return: current env
    """
    # Return value is presumably one of the above-declared environments
    return app.registry.settings.get('env.name')


def get_deployment_config(app):
    """
    Gets deployment configuration for the current environment.

    Sets ENV_NAME and WIPE_ES as side-effects.

    :param app: handle to Pyramid app
    :return: dict of config options
    """
    deploy_cfg = {}
    my_env = get_my_env(app)
    deploy_cfg['ENV_NAME'] = my_env
    if is_stg_or_prd_env(my_env):
        log.info('This looks like our production environment -- not wiping ES')
        deploy_cfg['WIPE_ES'] = False
    elif is_test_env(my_env):
        log.info('This looks like a test environment -- wiping ES')
        deploy_cfg['WIPE_ES'] = True
    else:
        log.info('This environment is not recognized -- not wiping ES')
        deploy_cfg['WIPE_ES'] = False
    return deploy_cfg


def _run_create_mapping(app, args):
    """
    Runs create_mapping with deploy options and report errors. Allows args passed from argparse in main to override
    the default deployment configuration

    :param app: pyramid application handle
    :param args: args from argparse
    :return: None
    """
    try:
        deploy_cfg = get_deployment_config(app)
        log.info('Running create mapping on env: %s' % deploy_cfg['ENV_NAME'])
        if args.wipe_es:  # override deploy_cfg WIPE_ES option
            log.info('Overriding deploy_cfg and wiping ES')
            deploy_cfg['WIPE_ES'] = True
        run_create_mapping(app, check_first=(not deploy_cfg['WIPE_ES']), purge_queue=args.clear_queue,
                           item_order=loadxl_order)
    except Exception as e:
        log.error("Exception encountered while gathering deployment information or running create_mapping")
        log.error(str(e))
        exit(1)


def main():
    parser = argparse.ArgumentParser(  # noqa - PyCharm wrongly thinks the formatter_class is invalid
        description="Create Elasticsearch mapping on deployment", epilog=EPILOG,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument('config_uri', help="path to configfile")
    parser.add_argument('--app-name', help="Pyramid app name in configfile")
    parser.add_argument('--wipe-es', help="Specify to wipe ES", action='store_true', default=False)
    parser.add_argument('--clear-queue', help="Specify to clear the SQS queue", action='store_true', default=False)
    parser.add_argument('--staggered', default=False, action='store_true',
                        help='Pass to trigger staggered reindexing, a new mode that will go type-by-type')

    args = parser.parse_args()
    app = get_app(args.config_uri, args.app_name)
    # Loading app will have configured from config file. Reconfigure here:
    set_logging(in_prod=app.registry.settings.get('production'), log_name=__name__, level=logging.DEBUG)
    # set_logging(app.registry.settings.get('elasticsearch.server'), app.registry.settings.get('production'),
    #             level=logging.DEBUG)
    if args.staggered:
        reindex_by_type_staggered(app)  # note that data flow from args is dropped, only 1 mode for staggered
    else:
        _run_create_mapping(app, args)
    exit(0)


if __name__ == '__main__':
    main()



# Authorization related functions which may be overriden by an implementing app,
# e.g. Foursight or CGAP portal, using the dcicutils project_utils mechanism.

from ..authorization import _create_principals

class SnovaultProjectAuthorization:

    def authorization_create_principals(self, login, user, collections):
        return _create_principals(login, user, collections)



import pytest


TEST_COLLECTION = '/testing_mixins'


@pytest.fixture
def test_object_with_mixin():
    return {
        'name': 'i_should_have_a_mixin'
    }


def test_mixins_basic(authenticated_testapp, test_object_with_mixin):
    resp = authenticated_testapp.post_json(TEST_COLLECTION, test_object_with_mixin, status=201)
    assert 'status' in resp.json['@graph'][0]



from dcicutils import creds_utils


def get_auth_key(portal: str, env: str) -> str:
    """Get the auth key for the given portal and environment."""
    key_manager = creds_utils.KeyManager.create(portal)
    return key_manager.get_keydict_for_env(env)



# --------------------------------------------------------------------------------------------------
# Command-line utility to retrieve and print the given object (UUID) from a locally running portal.
# --------------------------------------------------------------------------------------------------
# Example command:
#  view-local-object 4483b19d-62e7-4e7f-a211-0395343a35df
#
# Example output:
#  (smaht-portal-3.9.16) mac: view-local-object 3968e38e-c11f-472e-8531-8650e2e296d4 --yaml
#  '@context': /terms/
#  '@id': /access-keys/3968e38e-c11f-472e-8531-8650e2e296d4/
#  '@type':
#  - AccessKey
#  - Item
#  access_key_id: NSVCZ75O
#  actions:
#  - href: /access-keys/3968e38e-c11f-472e-8531-8650e2e296d4/?currentAction=create
#    name: create
#    profile: /profiles/AccessKey.json
#    title: Create
#  - href: /access-keys/3968e38e-c11f-472e-8531-8650e2e296d4/?currentAction=edit
#    name: edit
#    profile: /profiles/AccessKey.json
#    title: Edit
#  aggregated-items: {}
#  date_created: '2023-09-06T13:11:59.704005+00:00'
#  description: Manually generated local access-key for testing.
#  display_title: AccessKey from 2023-09-06
#  expiration_date: '2023-12-05T13:11:59.714106'
#  last_modified:
#    date_modified: '2023-09-06T13:11:59.711367+00:00'
#    modified_by:
#      '@id': /users/3202fd57-44d2-44fb-a131-afb1e43d8ae5/
#      '@type':
#      - User
#      - Item
#      display_title: loadxl loadxl
#      principals_allowed:
#        edit:
#        - group.admin
#        view:
#        - group.admin
# #         - group.read-only-admin
#        - remoteuser.EMBED
#        - remoteuser.INDEXER
#      status: current
#      uuid: 3202fd57-44d2-44fb-a131-afb1e43d8ae5
#  principals_allowed:
#    edit:
#    - group.admin
#    - userid.74fef71a-dfc1-4aa4-acc0-cedcb7ac1d68
#    view:
#    - group.admin
#    - group.read-only-admin
#    - remoteuser.EMBED
#    - remoteuser.INDEXER
#    - userid.74fef71a-dfc1-4aa4-acc0-cedcb7ac1d68
#  schema_version: '1'
#  status: current
#  user:
#    '@id': /users/74fef71a-dfc1-4aa4-acc0-cedcb7ac1d68/
#    '@type':
#    - User
#    - Item
#    display_title: David Michaels
#    principals_allowed:
#      edit:
#      - group.admin
#      view:
#      - group.admin
#      - group.read-only-admin
#      - remoteuser.EMBED
#      - remoteuser.INDEXER
#    status: current
#    uuid: 74fef71a-dfc1-4aa4-acc0-cedcb7ac1d68
#  uuid: 3968e38e-c11f-472e-8531-8650e2e296d4
#  validation-errors: []
#
# Note that instead of a uuid you can also actually use a path, for example:
#   view-local-object /file-formats/vcf_gz_tbi
#
# --------------------------------------------------------------------------------------------------

import argparse
import json
import sys
from typing import Optional
import yaml
from dcicutils.misc_utils import get_error_message
from dcicutils.portal_utils import Portal
from snovault.commands.captured_output import captured_output, uncaptured_output


_DEFAULT_INI_FILE = "development.ini"


def main():

    parser = argparse.ArgumentParser(description="Create local portal access-key for dev/testing purposes.")
    parser.add_argument("uuid", type=str,
                        help=f"The uuid (or path) of the object to fetch and view. ")
    parser.add_argument("--ini", type=str, required=False, default=None,
                        help=f"Name of the application .ini file; default is: {_DEFAULT_INI_FILE}")
    parser.add_argument("--env", "-e", type=str, required=False, default=None,
                        help=f"Environment name (key from ~/.smaht-keys.json).")
    parser.add_argument("--server", "-s", type=str, required=False, default=None,
                        help=f"Environment server name (server from key in ~/.smaht-keys.json).")
    parser.add_argument("--app", type=str, required=False, default=None,
                        help=f"Application name (one of: smaht, cgap, fourfront).")
    parser.add_argument("--raw", action="store_true", required=False, default=False, help="Raw output.")
    parser.add_argument("--database", action="store_true", required=False, default=False, help="Read from database output.")
    parser.add_argument("--yaml", action="store_true", required=False, default=False, help="YAML output.")
    parser.add_argument("--verbose", action="store_true", required=False, default=False, help="Verbose output.")
    parser.add_argument("--debug", action="store_true", required=False, default=False, help="Debugging output.")
    args = parser.parse_args()

    portal = _create_portal(ini=args.ini, env=args.env, server=args.server, app=args.app, debug=args.debug)
    data = _get_local_object(portal=portal, uuid=args.uuid, raw=args.raw, database=args.database, verbose=args.verbose)

    if args.yaml:
        _print(yaml.dump(data))
    else:
        _print(json.dumps(data, default=str, indent=4))


def _create_portal(ini: str = _DEFAULT_INI_FILE, env: Optional[str] = None,
                   server: Optional[str] = None, app: Optional[str] = None, debug: bool = False) -> dict:
    with captured_output(not debug):
        return Portal(env, server=server, app=app) if env or app else Portal(ini or _DEFAULT_INI_FILE)


def _get_local_object(portal: Portal, uuid: str,
                      raw: bool = False, database: bool = False, verbose: bool = False) -> dict:
    if verbose:
        _print(f"Getting object ({uuid}) from local portal ... ", end="")
    response = None
    try:
        if not uuid.startswith("/"):
            path = f"/{uuid}"
        else:
            path = uuid
        response = portal.get(path, raw=raw, database=database)
    except Exception as e:
        if "404" in str(e) and "not found" in str(e).lower():
            if verbose:
                _print("Not found!")
            else:
                _print(f"Object ({uuid}) not found!")
            _exit_without_action()
        _exit_without_action(f"Exception getting object ({uuid}) -> {get_error_message(e)}", newline=verbose)
    if not response:
        _exit_without_action(f"Null response getting object {uuid}).")
    if response.status_code not in [200, 307]:
        # TODO: Understand why the /me endpoin returns HTTP status code 307, which is only why we mention it above.
        _exit_without_action(f"Invalid status code ({response.status_code}) getting object: {uuid}")
    if not response.json:
        _exit_without_action(f"Invalid JSON getting object {uuid}).")
    if verbose:
        _print("OK")
    return response.json()


def _print(*args, **kwargs):
    with uncaptured_output():
        print(*args, **kwargs)
    sys.stdout.flush()


def _exit_without_action(message: Optional[str] = None, newline: bool = True) -> None:
    if message:
        if newline:
            _print()
        _print(f"ERROR: {message}")
    exit(1)


if __name__ == "__main__":
    main()
